# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import List, Optional  # noqa: F401

from databuilder.models.elasticsearch_document import ElasticsearchDocument


class DashboardESDocument(ElasticsearchDocument):
    """
    Schema for the ES dashboard ES document
    """

    def __init__(self,
                 group_name,  # type: str
                 name,  # type: str
                 description,  # type: Union[str, None]
                 total_usage,  # type: int
                 product='',  # type: Optional[str]
                 cluster='',  # type: Optional[str]
                 group_description=None,  # type: Optional[str]
                 query_names=None,  # type:  Union[List[str], None]
                 group_url=None,  # type: Optional[str]
                 url=None,  # type: Optional[str]
                 uri=None,  # type: Optional[str]
                 last_successful_run_timestamp=None,  # type: Optional[int]
                 tags=None,  # type: Optional[list[str]]
                 badges=None,  # type: Optional[list[str]]
                 ):
        # type: (...) -> None
        self.group_name = group_name
        self.name = name
        self.description = description
        self.cluster = cluster
        self.product = product
        self.group_url = group_url
        self.url = url
        self.uri = uri
        self.last_successful_run_timestamp = last_successful_run_timestamp
        self.total_usage = total_usage
        self.group_description = group_description
        self.query_names = query_names
        self.tags = tags
        self.badges = badges



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging

import requests  # noqa: F401
from jsonpath_rw import parse
from typing import Any  # noqa: F401

from databuilder.rest_api.rest_api_query import RestApiQuery

#  How many records considers as full and indicating there might be next page? In list reports on space API, it's 30.
DEFAULT_MAX_RECORD_SIZE = 30
PAGE_SUFFIX_TEMPLATE = '?page={}'
LIST_REPORTS_PAGINATION_JSON_PATH = '_embedded.reports[*]'  # So far this is the only paginated API that we need.

LOGGER = logging.getLogger(__name__)


class ModePaginatedRestApiQuery(RestApiQuery):
    """
    Certain API such as get list of reports on a space is paginated with query term page.
    https://mode.com/developer/api-cookbook/management/get-all-reports/

    This subclass makes sure to detect if there's more page and update URL to get next page.
    """

    def __init__(self,
                 pagination_json_path=LIST_REPORTS_PAGINATION_JSON_PATH,  # type: str
                 max_record_size=DEFAULT_MAX_RECORD_SIZE,  # type: int
                 **kwargs  # type: Any
                 ):
        # type (...) -> None
        super(ModePaginatedRestApiQuery, self).__init__(**kwargs)

        self._original_url = self._url
        self._max_record_size = max_record_size
        self._current_page = 1
        self._pagination_jsonpath_expr = parse(pagination_json_path)

    def _preprocess_url(self,
                        record,  # type: Dict[str, Any]
                        ):
        # type: (...) -> str
        """
        Updates URL with page information
        :param record:
        :return: a URL that is ready to be called.
        """
        page_suffix = PAGE_SUFFIX_TEMPLATE.format(self._current_page)  # example: ?page=2

        # example: http://foo.bar/resources?page=2
        self._url = self._original_url + '{page_suffix}'.format(original_url=self._original_url,
                                                                page_suffix=page_suffix)
        return self._url.format(**record)

    def _post_process(self,
                      response,  # type: requests.Response
                      ):
        # type: (...) -> None
        """
        Updates trigger to pagination (self._more_pages) as well as current_page (self._current_page)
        Mode does not have explicit indicator that it just the number of records need to be certain number that
        implying that there could be more records on next page.
        :return:
        """

        result_list = [match.value for match in self._pagination_jsonpath_expr.find(response.json())]

        if result_list and len(result_list) >= self._max_record_size:
            self._more_pages = True
            self._current_page = self._current_page + 1
            return

        self._more_pages = False
        self._current_page = 1



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import Dict, Any, Union, Iterator  # noqa: F401

from databuilder.models.neo4j_csv_serde import (
    Neo4jCsvSerializable, NODE_LABEL, NODE_KEY)
from databuilder.models.schema.schema_constant import SCHEMA_NODE_LABEL, SCHEMA_NAME_ATTR
from databuilder.models.table_metadata import DescriptionMetadata


class SchemaModel(Neo4jCsvSerializable):

    def __init__(self,
                 schema_key,
                 schema,
                 description=None,
                 description_source=None,
                 **kwargs):
        self._schema_key = schema_key
        self._schema = schema
        self._description = DescriptionMetadata.create_description_metadata(text=description,
                                                                            source=description_source) \
            if description else None
        self._node_iterator = self._create_node_iterator()
        self._relation_iterator = self._create_relation_iterator()

    def create_next_node(self):
        # type: () -> Union[Dict[str, Any], None]
        try:
            return next(self._node_iterator)
        except StopIteration:
            return None

    def _create_node_iterator(self):
        # type: () -> Iterator[[Dict[str, Any]]]
        yield {
            NODE_LABEL: SCHEMA_NODE_LABEL,
            NODE_KEY: self._schema_key,
            SCHEMA_NAME_ATTR: self._schema,
        }

        if self._description:
            yield self._description.get_node_dict(self._get_description_node_key())

    def create_next_relation(self):
        # type: () -> Union[Dict[str, Any], None]
        try:
            return next(self._relation_iterator)
        except StopIteration:
            return None

    def _get_description_node_key(self):
        return '{}/{}'.format(self._schema_key, self._description.get_description_id())

    def _create_relation_iterator(self):
        # type: () -> Iterator[[Dict[str, Any]]]
        if self._description:
            yield self._description.get_relation(start_node=SCHEMA_NODE_LABEL,
                                                 start_key=self._schema_key,
                                                 end_key=self._get_description_node_key())



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from pyhocon import ConfigTree  # noqa: F401
from typing import Any  # noqa: F401

from databuilder.transformer.base_transformer import Transformer


LOGGER = logging.getLogger(__name__)


# Config keys
REGEX_REPLACE_TUPLE_LIST = 'regex_replace_tuple_list'
ATTRIBUTE_NAME = 'attribute_name'


class RegexStrReplaceTransformer(Transformer):
    """
    Generic string replacement transformer using REGEX.
    User can pass list of tuples where tuple contains regex and replacement pair.

    Any non-string values will be ignored.
    """

    def init(self, conf):
        # type: (ConfigTree) -> None
        self._regex_replace_tuples = conf.get_list(REGEX_REPLACE_TUPLE_LIST)
        self._attribute_name = conf.get_string(ATTRIBUTE_NAME)

    def transform(self, record):
        # type: (Any) -> Any

        if isinstance(record, dict):
            val = record.get(self._attribute_name)
        else:
            val = getattr(record, self._attribute_name)

        if val is None or not isinstance(val, str):
            return record

        for regex_replace_tuple in self._regex_replace_tuples:
            val = val.replace(regex_replace_tuple[0], regex_replace_tuple[1])

        if isinstance(record, dict):
            record[self._attribute_name] = val
        else:
            setattr(record, self._attribute_name, val)

        return record

    def get_scope(self):
        # type: () -> str
        return 'transformer.regex_str_replace'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from cassandra.cluster import Cluster

from pyhocon import ConfigFactory, ConfigTree  # noqa: F401
from typing import Iterator, Union, Dict, Any  # noqa: F401

from databuilder.extractor.base_extractor import Extractor
from databuilder.models.table_metadata import TableMetadata, ColumnMetadata


class CassandraExtractor(Extractor):
    """
    Extracts tables and columns metadata from Apacha Cassandra
    """

    CLUSTER_KEY = 'cluster'
    # Key to define clusters ips, it should be List[str]
    IPS_KEY = 'ips'
    # Key to define extra kwargs to pass on cluster constructor,
    # it should be Dict[Any]
    KWARGS_KEY = 'kwargs'
    # Key to define custom filter function based on keyspace and table
    # since the cluster metadata doesn't support native filters,
    # it should be like def filter(keyspace, table) and return False if
    # going to skip that table and True if not
    FILTER_FUNCTION_KEY = 'filter'

    # Default values
    DEFAULT_CONFIG = ConfigFactory.from_dict({
        CLUSTER_KEY: 'gold',
        IPS_KEY: [],
        KWARGS_KEY: {},
        FILTER_FUNCTION_KEY: None
    })

    def init(self, conf):
        conf = conf.with_fallback(CassandraExtractor.DEFAULT_CONFIG)
        self._cluster = '{}'.format(conf.get_string(CassandraExtractor.CLUSTER_KEY))
        self._filter = conf.get(CassandraExtractor.FILTER_FUNCTION_KEY)
        ips = conf.get_list(CassandraExtractor.IPS_KEY)
        kwargs = conf.get(CassandraExtractor.KWARGS_KEY)
        self._client = Cluster(ips, **kwargs)
        self._client.connect()
        self._extract_iter = None  # type: Union[None, Iterator]

    def extract(self):
        # type: () -> Union[TableMetadata, None]
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def get_scope(self):
        # type: () -> str
        return 'extractor.cassandra'

    def _get_extract_iter(self):
        # type: () -> Iterator[TableMetadata]
        """
        It gets all tables and yields TableMetadata
        :return:
        """
        keyspaces = self._get_keyspaces()
        for keyspace in keyspaces:
            # system keyspaces
            if keyspace.startswith('system'):
                continue
            for table in self._get_tables(keyspace):
                if self._filter and not self._filter(keyspace, table):
                    continue

                columns = []

                columns_dict = self._get_columns(keyspace, table)
                for idx, (column_name, column) in enumerate(columns_dict.items()):
                    columns.append(ColumnMetadata(
                        column_name,
                        None,
                        column.cql_type,
                        idx
                    ))

                yield TableMetadata(
                    'cassandra',
                    self._cluster,
                    keyspace,
                    table,
                    None,
                    columns
                )

    def _get_keyspaces(self):
        return self._client.metadata.keyspaces

    def _get_tables(self, keyspace):
        return self._client.metadata.keyspaces[keyspace].tables

    def _get_columns(self, keyspace, table):
        return self._client.metadata.keyspaces[keyspace].tables[table].columns



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Dict, List, Union  # noqa: F401

from databuilder.models.neo4j_csv_serde import Neo4jCsvSerializable, NODE_KEY, \
    NODE_LABEL, RELATION_START_KEY, RELATION_START_LABEL, RELATION_END_KEY, \
    RELATION_END_LABEL, RELATION_TYPE, RELATION_REVERSE_TYPE
from databuilder.models.table_metadata import ColumnMetadata


class TableColumnStats(Neo4jCsvSerializable):
    # type: (...) -> None
    """
    Hive table stats model.
    Each instance represents one row of hive watermark result.
    """
    LABEL = 'Stat'
    KEY_FORMAT = '{db}://{cluster}.{schema}' \
                 '/{table}/{col}/{stat_name}/'
    STAT_Column_RELATION_TYPE = 'STAT_OF'
    Column_STAT_RELATION_TYPE = 'STAT'

    def __init__(self,
                 table_name,  # type: str
                 col_name,  # type: str
                 stat_name,  # type: str
                 stat_val,  # type: str
                 start_epoch,  # type: str
                 end_epoch,  # type: str
                 db='hive',  # type: str
                 cluster='gold',  # type: str
                 schema=None  # type: str
                 ):
        # type: (...) -> None
        if schema is None:
            self.schema, self.table = table_name.split('.')
        else:
            self.table = table_name.lower()
            self.schema = schema.lower()
        self.db = db
        self.col_name = col_name.lower()
        self.start_epoch = start_epoch
        self.end_epoch = end_epoch
        self.cluster = cluster
        self.stat_name = stat_name
        self.stat_val = stat_val
        self._node_iter = iter(self.create_nodes())
        self._relation_iter = iter(self.create_relation())

    def create_next_node(self):
        # type: (...) -> Union[Dict[str, Any], None]
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self):
        # type: (...) -> Union[Dict[str, Any], None]
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def get_table_stat_model_key(self):
        # type: (...) -> str
        return TableColumnStats.KEY_FORMAT.format(db=self.db,
                                                  cluster=self.cluster,
                                                  schema=self.schema,
                                                  table=self.table,
                                                  col=self.col_name,
                                                  stat_name=self.stat_name)

    def get_col_key(self):
        # type: (...) -> str
        # no cluster, schema info from the input
        return ColumnMetadata.COLUMN_KEY_FORMAT.format(db=self.db,
                                                       cluster=self.cluster,
                                                       schema=self.schema,
                                                       tbl=self.table,
                                                       col=self.col_name)

    def create_nodes(self):
        # type: () -> List[Dict[str, Any]]
        """
        Create a list of Neo4j node records
        :return:
        """
        results = [{
            NODE_KEY: self.get_table_stat_model_key(),
            NODE_LABEL: TableColumnStats.LABEL,
            'stat_val:UNQUOTED': self.stat_val,
            'stat_name': self.stat_name,
            'start_epoch': self.start_epoch,
            'end_epoch': self.end_epoch,
        }]
        return results

    def create_relation(self):
        # type: () -> List[Dict[str, Any]]
        """
        Create a list of relation map between table stat record with original hive table
        :return:
        """
        results = [{
            RELATION_START_KEY: self.get_table_stat_model_key(),
            RELATION_START_LABEL: TableColumnStats.LABEL,
            RELATION_END_KEY: self.get_col_key(),
            RELATION_END_LABEL: ColumnMetadata.COLUMN_NODE_LABEL,
            RELATION_TYPE: TableColumnStats.STAT_Column_RELATION_TYPE,
            RELATION_REVERSE_TYPE: TableColumnStats.Column_STAT_RELATION_TYPE
        }]
        return results



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from pyhocon import ConfigTree, ConfigFactory  # noqa: F401
from requests.auth import HTTPBasicAuth

from databuilder import Scoped
from databuilder.extractor.dashboard.mode_analytics.mode_dashboard_constants import ORGANIZATION, MODE_ACCESS_TOKEN, \
    MODE_PASSWORD_TOKEN
from databuilder.extractor.restapi.rest_api_extractor import RestAPIExtractor, REST_API_QUERY, STATIC_RECORD_DICT
from databuilder.rest_api.base_rest_api_query import BaseRestApiQuery  # noqa: F401
from databuilder.rest_api.base_rest_api_query import RestApiQuerySeed
from databuilder.rest_api.rest_api_query import RestApiQuery  # noqa: F401


class ModeDashboardUtils(object):

    @staticmethod
    def get_spaces_query_api(conf,  # type: ConfigTree
                             ):
        """
        Provides RestApiQuerySeed where it will provides iterator of dictionaries as records where dictionary keys are
         organization, dashboard_group_id, dashboard_group and dashboard_group_description
        :param conf:
        :return:
        """
        # type: (...) -> BaseRestApiQuery

        # https://mode.com/developer/api-reference/management/spaces/#listSpaces
        spaces_url_template = 'https://app.mode.com/api/{organization}/spaces?filter=all'

        # Seed query record for next query api to join with
        seed_record = [{'organization': conf.get_string(ORGANIZATION)}]
        seed_query = RestApiQuerySeed(seed_record=seed_record)

        # Spaces
        params = {'auth': HTTPBasicAuth(conf.get_string(MODE_ACCESS_TOKEN),
                                        conf.get_string(MODE_PASSWORD_TOKEN))}

        json_path = '_embedded.spaces[*].[token,name,description]'
        field_names = ['dashboard_group_id', 'dashboard_group', 'dashboard_group_description']
        spaces_query = RestApiQuery(query_to_join=seed_query, url=spaces_url_template, params=params,
                                    json_path=json_path, field_names=field_names)

        return spaces_query

    @staticmethod
    def get_auth_params(conf,  # type: ConfigTree
                        ):
        params = {'auth': HTTPBasicAuth(conf.get_string(MODE_ACCESS_TOKEN),
                                        conf.get_string(MODE_PASSWORD_TOKEN)
                                        )
                  }
        return params

    @staticmethod
    def create_mode_rest_api_extractor(restapi_query,  # type: BaseRestApiQuery
                                       conf,  # type: ConfigTree
                                       ):
        """
        Creates RestAPIExtractor. Note that RestAPIExtractor is already initialized
        :param restapi_query:
        :param conf:
        :return: RestAPIExtractor. Note that RestAPIExtractor is already initialized
        """
        extractor = RestAPIExtractor()
        rest_api_extractor_conf = \
            Scoped.get_scoped_conf(conf, extractor.get_scope())\
                  .with_fallback(conf)\
                  .with_fallback(ConfigFactory.from_dict({REST_API_QUERY: restapi_query,
                                                          STATIC_RECORD_DICT: {'product': 'mode'}
                                                          }
                                                         )
                                 )

        extractor.init(conf=rest_api_extractor_conf)
        return extractor



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import importlib
from pyhocon import ConfigFactory
from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata
from databuilder.models.dashboard.dashboard_last_modified import DashboardLastModifiedTimestamp
from databuilder.models.dashboard.dashboard_owner import DashboardOwner
from databuilder.models.dashboard.dashboard_query import DashboardQuery
from databuilder.models.dashboard.dashboard_table import DashboardTable
from databuilder.models.table_metadata import TableMetadata
from databuilder.extractor.base_extractor import Extractor
from databuilder.rest_api.rest_api_query import RestApiQuery
from databuilder.rest_api.base_rest_api_query import EmptyRestApiQuerySeed
from databuilder.extractor.restapi.rest_api_extractor import RestAPIExtractor, REST_API_QUERY
from databuilder.extractor.dashboard.redash.redash_dashboard_utils import \
    get_auth_headers, get_text_widgets, get_visualization_widgets, sort_widgets, \
    generate_dashboard_description, RedashPaginatedRestApiQuery
from databuilder.transformer.base_transformer import ChainedTransformer
from databuilder.transformer.timestamp_string_to_epoch import TimestampStringToEpoch, FIELD_NAME as TS_FIELD_NAME


class TableRelationData:
    """
    This is sort of like a stripped down version of `TableMetadata`.
    It is used as the type returned by the (optional) table parser.
    """

    def __init__(self, database, cluster, schema, name):
        # type: (str, str, str, str) -> None

        self._data = {'db': database, 'cluster': cluster, 'schema': schema, 'tbl': name}

    @property
    def key(self):
        # type: () -> str

        return TableMetadata.TABLE_KEY_FORMAT.format(**self._data)


class RedashDashboardExtractor(Extractor):
    """
    An extractor for retrieving dashboards and associated queries
    (and possibly tables) from Redash.

    There are five configuration values:

    - `redash_base_url`: (e.g., `https://redash.example.com`) Base URL for the user-facing
    Redash application
    - `api_base_url`: (e.g., `https://redash.example.com/api`) Base URL for the API
    - `api_key`: Redash API key
    - (optional) `cluster`: A cluster name for this Redash instance (defaults to `prod`)
    - (optional) `table_parser`: A function `(RedashVisualizationWidget) -> List[TableRelationData]`.
    Given a `RedashVisualizationWidget`, this should return a list of potentially related tables
    in Amundsen. Any table returned that exists in Amundsen will be linked to the dashboard.
    Any table that does not exist will be ignored.
    """

    REDASH_BASE_URL_KEY = 'redash_base_url'
    API_BASE_URL_KEY = 'api_base_url'
    API_KEY_KEY = 'api_key'
    CLUSTER_KEY = 'cluster'  # optional config
    TABLE_PARSER_KEY = 'table_parser'  # optional config

    DEFAULT_CLUSTER = 'prod'

    PRODUCT = 'redash'
    DASHBOARD_GROUP_ID = 'redash'
    DASHBOARD_GROUP_NAME = 'Redash'

    def init(self, conf):
        # type: (ConfigTree) -> None

        # required configuration
        self._redash_base_url = conf.get_string(RedashDashboardExtractor.REDASH_BASE_URL_KEY)
        self._api_base_url = conf.get_string(RedashDashboardExtractor.API_BASE_URL_KEY)
        self._api_key = conf.get_string(RedashDashboardExtractor.API_KEY_KEY)

        # optional configuration
        self._cluster = conf.get_string(
            RedashDashboardExtractor.CLUSTER_KEY, RedashDashboardExtractor.DEFAULT_CLUSTER
        )
        self._parse_tables = None
        tbl_parser_path = conf.get_string(RedashDashboardExtractor.TABLE_PARSER_KEY)
        if tbl_parser_path:
            module_name, fn_name = tbl_parser_path.rsplit('.', 1)
            mod = importlib.import_module(module_name)
            self._parse_tables = getattr(mod, fn_name)

        self._extractor = self._build_extractor()
        self._transformer = self._build_transformer()
        self._extract_iter = None

    def _is_published_dashboard(self, record):
        # type: Dict[str, Any] -> bool

        return not (record['is_archived'] or record['is_draft'])

    def _get_extract_iter(self):
        # type: () -> Iterator[Any]

        while True:
            record = self._extractor.extract()
            if not record:
                break  # the end.

            record = self._transformer.transform(record=record)

            if not self._is_published_dashboard(record):
                continue  # filter this one out

            identity_data = {
                'cluster': self._cluster,
                'product': RedashDashboardExtractor.PRODUCT,
                'dashboard_group_id': RedashDashboardExtractor.DASHBOARD_GROUP_ID,
                'dashboard_id': record['dashboard_id']
            }

            dash_data = {
                'dashboard_group':
                    RedashDashboardExtractor.DASHBOARD_GROUP_NAME,
                'dashboard_group_url':
                    self._redash_base_url,
                'dashboard_name':
                    record['dashboard_name'],
                'dashboard_url':
                    '{redash}/dashboard/{slug}'
                    .format(redash=self._redash_base_url, slug=record['slug']),
                'created_timestamp':
                    record['created_timestamp']
            }
            dash_data.update(identity_data)

            widgets = sort_widgets(record['widgets'])
            text_widgets = get_text_widgets(widgets)
            viz_widgets = get_visualization_widgets(widgets)

            # generate a description for this dashboard, since Redash does not have descriptions
            dash_data['description'] = generate_dashboard_description(text_widgets, viz_widgets)

            yield DashboardMetadata(**dash_data)

            last_mod_data = {'last_modified_timestamp': record['last_modified_timestamp']}
            last_mod_data.update(identity_data)

            yield DashboardLastModifiedTimestamp(**last_mod_data)

            owner_data = {'email': record['user']['email']}
            owner_data.update(identity_data)

            yield DashboardOwner(**owner_data)

            table_keys = set()

            for viz in viz_widgets:
                query_data = {
                    'query_id': viz.query_id,
                    'query_name': viz.query_name,
                    'url': self._redash_base_url + viz.query_relative_url,
                    'query_text': viz.raw_query
                }

                query_data.update(identity_data)
                yield DashboardQuery(**query_data)

                # if a table parser is provided, retrieve tables from this viz
                if self._parse_tables:
                    for tbl in self._parse_tables(viz):
                        table_keys.add(tbl.key)

            if len(table_keys) > 0:
                yield DashboardTable(table_ids=list(table_keys), **identity_data)

    def extract(self):
        # type: () -> Any

        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def _build_restapi_query(self):
        # type: () -> RestApiQuery

        dashes_query = RedashPaginatedRestApiQuery(
            query_to_join=EmptyRestApiQuerySeed(),
            url='{redash_api}/dashboards'.format(redash_api=self._api_base_url),
            params=self._get_default_api_query_params(),
            json_path='results[*].[id,name,slug,created_at,updated_at,is_archived,is_draft,user]',
            field_names=[
                'dashboard_id', 'dashboard_name', 'slug', 'created_timestamp',
                'last_modified_timestamp', 'is_archived', 'is_draft', 'user'
            ],
            skip_no_result=True
        )

        return RestApiQuery(
            query_to_join=dashes_query,
            url='{redash_api}/dashboards/{{slug}}'.format(redash_api=self._api_base_url),
            params=self._get_default_api_query_params(),
            json_path='widgets',
            field_names=['widgets'],
            skip_no_result=True
        )

    def _get_default_api_query_params(self):
        # type: () -> Dict[str, Any]

        return {'headers': get_auth_headers(self._api_key)}

    def _build_extractor(self):
        # type: () -> RestAPIExtractor

        extractor = RestAPIExtractor()
        rest_api_extractor_conf = ConfigFactory.from_dict({
            REST_API_QUERY: self._build_restapi_query()
        })
        extractor.init(rest_api_extractor_conf)
        return extractor

    def _build_transformer(self):
        # type: () -> ChainedTransformer

        transformers = []

        # transform timestamps from ISO to unix epoch
        ts_transformer_1 = TimestampStringToEpoch()
        ts_transformer_1.init(ConfigFactory.from_dict({
            TS_FIELD_NAME: 'created_timestamp',
        }))
        transformers.append(ts_transformer_1)

        ts_transformer_2 = TimestampStringToEpoch()
        ts_transformer_2.init(ConfigFactory.from_dict({
            TS_FIELD_NAME: 'last_modified_timestamp',
        }))
        transformers.append(ts_transformer_2)

        return ChainedTransformer(transformers=transformers)

    def get_scope(self):
        # type: () -> str

        return 'extractor.redash_dashboard'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import importlib
from sqlalchemy import create_engine

from pyhocon import ConfigTree  # noqa: F401
from typing import Any, Iterator  # noqa: F401

from databuilder.extractor.base_extractor import Extractor


class SQLAlchemyExtractor(Extractor):
    # Config keys
    CONN_STRING = 'conn_string'
    EXTRACT_SQL = 'extract_sql'
    """
    An Extractor that extracts records via SQLAlchemy. Database that supports SQLAlchemy can use this extractor
    """

    def init(self, conf):
        # type: (ConfigTree) -> None
        """
        Establish connections and import data model class if provided
        :param conf:
        """
        self.conf = conf
        self.conn_string = conf.get_string(SQLAlchemyExtractor.CONN_STRING)
        self.connection = self._get_connection()

        self.extract_sql = conf.get_string(SQLAlchemyExtractor.EXTRACT_SQL)

        model_class = conf.get('model_class', None)
        if model_class:
            module_name, class_name = model_class.rsplit(".", 1)
            mod = importlib.import_module(module_name)
            self.model_class = getattr(mod, class_name)
        self._execute_query()

    def _get_connection(self):
        # type: () -> Any
        """
        Create a SQLAlchemy connection to Database
        """
        engine = create_engine(self.conn_string)
        conn = engine.connect()
        return conn

    def _execute_query(self):
        # type: () -> None
        """
        Create an iterator to execute sql.
        """
        if not hasattr(self, 'results'):
            self.results = self.connection.execute(self.extract_sql)

        if hasattr(self, 'model_class'):
            results = [self.model_class(**result)
                       for result in self.results]
        else:
            results = self.results
        self.iter = iter(results)

    def extract(self):
        # type: () -> Any
        """
        Yield the sql result one at a time.
        convert the result to model if a model_class is provided
        """
        try:
            return next(self.iter)
        except StopIteration:
            return None
        except Exception as e:
            raise e

    def get_scope(self):
        # type: () -> str
        return 'extractor.sqlalchemy'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from collections import namedtuple

from typing import Iterable, Any, Union, Iterator, Dict, Set  # noqa: F401

# TODO: We could separate TagMetadata from table_metadata to own module
from databuilder.models.table_metadata import TagMetadata
from databuilder.models.neo4j_csv_serde import (
    Neo4jCsvSerializable, NODE_LABEL, NODE_KEY, RELATION_START_KEY, RELATION_END_KEY, RELATION_START_LABEL,
    RELATION_END_LABEL, RELATION_TYPE, RELATION_REVERSE_TYPE)


NodeTuple = namedtuple('KeyName', ['key', 'name', 'label'])
RelTuple = namedtuple('RelKeys', ['start_label', 'end_label', 'start_key', 'end_key', 'type', 'reverse_type'])


class MetricMetadata(Neo4jCsvSerializable):
    """
    Table metadata that contains columns. It implements Neo4jCsvSerializable so that it can be serialized to produce
    Table, Column and relation of those along with relationship with table and schema. Additionally, it will create
    Database, Cluster, and Schema with relastionships between those.
    These are being created here as it does not make much sense to have different extraction to produce this. As
    database, cluster, schema would be very repititive with low cardinality, it will perform de-dupe so that publisher
    won't need to publish same nodes, relationships.

    This class can be used for both table and view metadata. If it is a View, is_view=True should be passed in.
    """

    DESCRIPTION_NODE_LABEL = 'Description'

    METRIC_NODE_LABEL = 'Metric'
    METRIC_KEY_FORMAT = 'metric://{name}'
    METRIC_NAME = 'name'

    METRIC_DESCRIPTION = 'description'
    METRIC_DESCRIPTION_FORMAT = 'metric://{name}/_description'
    METRIC_DESCRIPTION_RELATION_TYPE = 'DESCRIPTION'
    DESCRIPTION_METRIC_RELATION_TYPE = 'DESCRIPTION_OF'

    DASHBOARD_NODE_LABEL = 'Dashboard'
    DASHBOARD_KEY_FORMAT = '{dashboard_group}://{dashboard_name}'
    DASHBOARD_NAME = 'name'
    DASHBOARD_METRIC_RELATION_TYPE = 'METRIC'
    METRIC_DASHBOARD_RELATION_TYPE = 'METRIC_OF'

    METRIC_TYPE_NODE_LABEL = 'Metrictype'
    METRIC_TYPE_KEY_FORMAT = 'type://{type}'
    METRIC_METRIC_TYPE_RELATION_TYPE = 'METRIC_TYPE'
    METRIC_TYPE_METRIC_RELATION_TYPE = 'METRIC_TYPE_OF'

    # TODO: Idea, maybe move expression from attribute
    # to node or delete commented code below?
    # METRIC_EXPRESSION_NODE_LABEL = 'Metricexpression'
    # METRIC_EXPRESSION_KEY_FORMAT = 'expression://{name}'
    # METRIC_METRIC_EXPRESSION_RELATION_TYPE = 'DEFINITION'
    # METRIC_EXPRESSION_METRIC_RELATION_TYPE = 'DEFINITION_OF'
    METRIC_EXPRESSION_VALUE = 'expression'

    METRIC_TAG_RELATION_TYPE = 'TAG'
    TAG_METRIC_RELATION_TYPE = 'TAG_OF'

    serialized_nodes = set()  # type: Set[Any]
    serialized_rels = set()  # type: Set[Any]

    def __init__(self,
                 dashboard_group,  # type: str
                 dashboard_name,  # type: str
                 name,  # type: Union[str, None]
                 expression,   # type: str
                 description,   # type: str
                 type,   # type: str
                 tags,   # type: List
                 ):
        # type: (...) -> None

        self.dashboard_group = dashboard_group
        self.dashboard_name = dashboard_name
        self.name = name
        self.expression = expression
        self.description = description
        self.type = type
        self.tags = tags
        self._node_iterator = self._create_next_node()
        self._relation_iterator = self._create_next_relation()

    def __repr__(self):
        # type: () -> str
        return 'MetricMetadata({!r}, {!r}, {!r}, {!r}, {!r}, {!r}, {!r}'.format(
            self.dashboard_group,
            self.dashboard_name,
            self.name,
            self.expression,
            self.description,
            self.type,
            self.tags
        )

    def _get_metric_key(self):
        # type: () -> str
        return MetricMetadata.METRIC_KEY_FORMAT.format(name=self.name)

    def _get_metric_type_key(self):
        # type: () -> str
        return MetricMetadata.METRIC_TYPE_KEY_FORMAT.format(type=self.type)

    def _get_dashboard_key(self):
        # type: () -> str
        return MetricMetadata.DASHBOARD_KEY_FORMAT.format(dashboard_group=self.dashboard_group,
                                                          dashboard_name=self.dashboard_name)

    def _get_metric_description_key(self):
        # type: () -> str
        return MetricMetadata.METRIC_DESCRIPTION_FORMAT.format(name=self.name)

    def _get_metric_expression_key(self):
        # type: () -> str
        return MetricMetadata.METRIC_EXPRESSION_KEY_FORMAT.format(name=self.name)

    def create_next_node(self):
        # type: () -> Union[Dict[str, Any], None]
        try:
            return next(self._node_iterator)
        except StopIteration:
            return None

    def _create_next_node(self):
        # type: () -> Iterator[Any]

        # Metric node
        yield {NODE_LABEL: MetricMetadata.METRIC_NODE_LABEL,
               NODE_KEY: self._get_metric_key(),
               MetricMetadata.METRIC_NAME: self.name,
               MetricMetadata.METRIC_EXPRESSION_VALUE: self.expression
               }

        # Description node
        if self.description:
            yield {NODE_LABEL: MetricMetadata.DESCRIPTION_NODE_LABEL,
                   NODE_KEY: self._get_metric_description_key(),
                   MetricMetadata.METRIC_DESCRIPTION: self.description}

        # Metric tag node
        if self.tags:
            for tag in self.tags:
                yield {NODE_LABEL: TagMetadata.TAG_NODE_LABEL,
                       NODE_KEY: TagMetadata.get_tag_key(tag),
                       TagMetadata.TAG_TYPE: 'metric'}

        # Metric type node
        if self.type:
            yield {NODE_LABEL: MetricMetadata.METRIC_TYPE_NODE_LABEL,
                   NODE_KEY: self._get_metric_type_key(),
                   'name': self.type}

        others = []

        for node_tuple in others:
            if node_tuple not in MetricMetadata.serialized_nodes:
                MetricMetadata.serialized_nodes.add(node_tuple)
                yield {
                    NODE_LABEL: node_tuple.label,
                    NODE_KEY: node_tuple.key,
                    'name': node_tuple.name
                }

    def create_next_relation(self):
        # type: () -> Union[Dict[str, Any], None]
        try:
            return next(self._relation_iterator)
        except StopIteration:
            return None

    def _create_next_relation(self):
        # type: () -> Iterator[Any]

        # Dashboard > Metric relation
        yield {
            RELATION_START_LABEL: MetricMetadata.METRIC_NODE_LABEL,
            RELATION_END_LABEL: MetricMetadata.DASHBOARD_NODE_LABEL,
            RELATION_START_KEY: self._get_metric_key(),
            RELATION_END_KEY: self._get_dashboard_key(),
            RELATION_TYPE: MetricMetadata.METRIC_DASHBOARD_RELATION_TYPE,
            RELATION_REVERSE_TYPE: MetricMetadata.DASHBOARD_METRIC_RELATION_TYPE
        }

        # Metric > Metric description relation
        if self.description:
            yield {
                RELATION_START_LABEL: MetricMetadata.METRIC_NODE_LABEL,
                RELATION_END_LABEL: MetricMetadata.DESCRIPTION_NODE_LABEL,
                RELATION_START_KEY: self._get_metric_key(),
                RELATION_END_KEY: self._get_metric_description_key(),
                RELATION_TYPE: MetricMetadata.METRIC_DESCRIPTION_RELATION_TYPE,
                RELATION_REVERSE_TYPE: MetricMetadata.DESCRIPTION_METRIC_RELATION_TYPE
            }

        # Metric > Metric tag relation
        if self.tags:
            for tag in self.tags:
                yield {
                    RELATION_START_LABEL: MetricMetadata.METRIC_NODE_LABEL,
                    RELATION_END_LABEL: TagMetadata.TAG_NODE_LABEL,
                    RELATION_START_KEY: self._get_metric_key(),
                    RELATION_END_KEY: TagMetadata.get_tag_key(tag),
                    RELATION_TYPE: MetricMetadata.METRIC_TAG_RELATION_TYPE,
                    RELATION_REVERSE_TYPE: MetricMetadata.TAG_METRIC_RELATION_TYPE
                }

        # Metric > Metric type relation
        if self.type:
            yield {
                RELATION_START_LABEL: MetricMetadata.METRIC_NODE_LABEL,
                RELATION_END_LABEL: MetricMetadata.METRIC_TYPE_NODE_LABEL,
                RELATION_START_KEY: self._get_metric_key(),
                RELATION_END_KEY: self._get_metric_type_key(),
                RELATION_TYPE: MetricMetadata.METRIC_METRIC_TYPE_RELATION_TYPE,
                RELATION_REVERSE_TYPE: MetricMetadata.METRIC_TYPE_METRIC_RELATION_TYPE
            }

        others = []

        for rel_tuple in others:
            if rel_tuple not in MetricMetadata.serialized_rels:
                MetricMetadata.serialized_rels.add(rel_tuple)
                yield {
                    RELATION_START_LABEL: rel_tuple.start_label,
                    RELATION_END_LABEL: rel_tuple.end_label,
                    RELATION_START_KEY: rel_tuple.start_key,
                    RELATION_END_KEY: rel_tuple.end_key,
                    RELATION_TYPE: rel_tuple.type,
                    RELATION_REVERSE_TYPE: rel_tuple.reverse_type
                }



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import abc

from pyhocon import ConfigTree  # noqa: F401
from typing import Any  # noqa: F401

from databuilder import Scoped


class Extractor(Scoped):
    """
    An extractor extracts record
    """

    @abc.abstractmethod
    def init(self, conf):
        # type: (ConfigTree) -> None
        pass

    @abc.abstractmethod
    def extract(self):
        # type: () -> Any
        """
        :return: Provides a record or None if no more to extract
        """
        return None

    def get_scope(self):
        # type: () -> str
        return 'extractor'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Any, Dict  # noqa: F401

from pyhocon import ConfigTree  # noqa: F401

from databuilder.transformer.base_transformer import Transformer

FIELD_NAMES = 'field_names'  # field name to be removed

LOGGER = logging.getLogger(__name__)


class RemoveFieldTransformer(Transformer):
    """
    Remove field in Dict by specifying list of fields (keys).

    """

    def init(self, conf):
        # type: (ConfigTree) -> None
        self._field_names = conf.get_list(FIELD_NAMES)

    def transform(self, record):
        # type: (Dict[str, Any]) -> Dict[str, Any]

        for k in self._field_names:
            if k in record:
                del record[k]

        return record

    def get_scope(self):
        # type: () -> str
        return 'transformer.remove_field'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging

from pyhocon import ConfigTree, ConfigFactory  # noqa: F401
from typing import Any  # noqa: F401

from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.dashboard.mode_analytics.mode_dashboard_utils import ModeDashboardUtils
from databuilder.extractor.restapi.rest_api_extractor import MODEL_CLASS
from databuilder.rest_api.mode_analytics.mode_paginated_rest_api_query import ModePaginatedRestApiQuery
from databuilder.rest_api.rest_api_failure_handlers import HttpFailureSkipOnStatus
from databuilder.rest_api.rest_api_query import RestApiQuery

LOGGER = logging.getLogger(__name__)


class ModeDashboardOwnerExtractor(Extractor):
    """
    An Extractor that extracts Dashboard owner.

    """

    def init(self, conf):
        # type: (ConfigTree) -> None
        self._conf = conf

        restapi_query = self._build_restapi_query()
        self._extractor = ModeDashboardUtils.create_mode_rest_api_extractor(
            restapi_query=restapi_query,
            conf=self._conf.with_fallback(
                ConfigFactory.from_dict(
                    {MODEL_CLASS: 'databuilder.models.dashboard.dashboard_owner.DashboardOwner', }
                )
            )
        )

    def extract(self):
        # type: () -> Any

        return self._extractor.extract()

    def get_scope(self):
        # type: () -> str
        return 'extractor.mode_dashboard_owner'

    def _build_restapi_query(self):
        """
        Build REST API Query. To get Mode Dashboard owner, it needs to call three APIs (spaces API, reports
        API, and user API) joining together.
        :return: A RestApiQuery that provides Mode Dashboard owner
        """
        # type: () -> RestApiQuery

        # https://mode.com/developer/api-reference/analytics/reports/#listReportsInSpace
        report_url_template = 'https://app.mode.com/api/{organization}/spaces/{dashboard_group_id}/reports'

        # https://mode.com/developer/api-reference/management/users/
        creator_url_template = 'https://app.mode.com{creator_resource_path}'

        spaces_query = ModeDashboardUtils.get_spaces_query_api(conf=self._conf)
        params = ModeDashboardUtils.get_auth_params(conf=self._conf)

        # Reports
        json_path = '(_embedded.reports[*].token) | (_embedded.reports[*]._links.creator.href)'
        field_names = ['dashboard_id', 'creator_resource_path']
        creator_resource_path_query = ModePaginatedRestApiQuery(query_to_join=spaces_query, url=report_url_template,
                                                                params=params,
                                                                json_path=json_path, field_names=field_names,
                                                                skip_no_result=True,
                                                                json_path_contains_or=True)

        json_path = 'email'
        field_names = ['email']
        failure_handler = HttpFailureSkipOnStatus(status_codes_to_skip={404})
        owner_email_query = RestApiQuery(query_to_join=creator_resource_path_query, url=creator_url_template,
                                         params=params,
                                         json_path=json_path, field_names=field_names, skip_no_result=True,
                                         can_skip_failure=failure_handler.can_skip_failure)

        return owner_email_query



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

ORGANIZATION = 'organization'
MODE_ACCESS_TOKEN = 'mode_user_token'
MODE_PASSWORD_TOKEN = 'mode_password_token'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Any  # noqa: F401

from pyhocon import ConfigTree, ConfigFactory  # noqa: F401
from requests.auth import HTTPBasicAuth

from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.dashboard.mode_analytics.mode_dashboard_constants import ORGANIZATION, MODE_ACCESS_TOKEN, \
    MODE_PASSWORD_TOKEN
from databuilder.extractor.dashboard.mode_analytics.mode_dashboard_utils import ModeDashboardUtils
from databuilder.rest_api.base_rest_api_query import RestApiQuerySeed
from databuilder.rest_api.rest_api_failure_handlers import HttpFailureSkipOnStatus
from databuilder.rest_api.rest_api_query import RestApiQuery
from databuilder.transformer.base_transformer import ChainedTransformer
from databuilder.transformer.dict_to_model import DictToModel, MODEL_CLASS
from databuilder.transformer.remove_field_transformer import RemoveFieldTransformer, FIELD_NAMES

LOGGER = logging.getLogger(__name__)


class ModeDashboardUserExtractor(Extractor):
    """
    An Extractor that extracts all Mode Dashboard user and add mode_user_id attribute to User model.
    """

    def init(self, conf):
        # type: (ConfigTree) -> None
        self._conf = conf

        restapi_query = self._build_restapi_query()
        self._extractor = ModeDashboardUtils.create_mode_rest_api_extractor(
            restapi_query=restapi_query,
            conf=self._conf
        )

        # Remove all unnecessary fields because User model accepts all attributes and push it to Neo4j.
        transformers = []

        remove_fields_transformer = RemoveFieldTransformer()
        remove_fields_transformer.init(
            conf=Scoped.get_scoped_conf(self._conf, remove_fields_transformer.get_scope()).with_fallback(
                ConfigFactory.from_dict(
                    {FIELD_NAMES: ['organization', 'mode_user_resource_path', 'product']})))
        transformers.append(remove_fields_transformer)

        dict_to_model_transformer = DictToModel()
        dict_to_model_transformer.init(
            conf=Scoped.get_scoped_conf(self._conf, dict_to_model_transformer.get_scope()).with_fallback(
                ConfigFactory.from_dict(
                    {MODEL_CLASS: 'databuilder.models.user.User'})))
        transformers.append(dict_to_model_transformer)

        self._transformer = ChainedTransformer(transformers=transformers)

    def extract(self):
        # type: () -> Any

        record = self._extractor.extract()
        if not record:
            return None

        return self._transformer.transform(record=record)

    def get_scope(self):
        # type: () -> str
        return 'extractor.mode_dashboard_owner'

    def _build_restapi_query(self):
        """
        Build REST API Query. To get Mode Dashboard owner, it needs to call three APIs (spaces API, reports
        API, and user API) joining together.
        :return: A RestApiQuery that provides Mode Dashboard owner
        """
        # type: () -> RestApiQuery

        # Seed query record for next query api to join with
        seed_record = [{
            'organization': self._conf.get_string(ORGANIZATION),
            'is_active': None,
            'updated_at': None,
            'do_not_update_empty_attribute': True,
        }]
        seed_query = RestApiQuerySeed(seed_record=seed_record)

        # memberships
        # https://mode.com/developer/api-reference/management/organization-memberships/#listMemberships
        memberships_url_template = 'https://app.mode.com/api/{organization}/memberships'
        params = {'auth': HTTPBasicAuth(self._conf.get_string(MODE_ACCESS_TOKEN),
                                        self._conf.get_string(MODE_PASSWORD_TOKEN))}

        json_path = '(_embedded.memberships[*].member_username) | (_embedded.memberships[*]._links.user.href)'
        field_names = ['mode_user_id', 'mode_user_resource_path']
        mode_user_ids_query = RestApiQuery(query_to_join=seed_query, url=memberships_url_template, params=params,
                                           json_path=json_path, field_names=field_names,
                                           skip_no_result=True, json_path_contains_or=True)

        # https://mode.com/developer/api-reference/management/users/
        user_url_template = 'https://app.mode.com{mode_user_resource_path}'

        json_path = 'email'
        field_names = ['email']
        failure_handler = HttpFailureSkipOnStatus(status_codes_to_skip={404})
        mode_user_email_query = RestApiQuery(query_to_join=mode_user_ids_query, url=user_url_template,
                                             params=params, json_path=json_path, field_names=field_names,
                                             skip_no_result=True, can_skip_failure=failure_handler.can_skip_failure)

        return mode_user_email_query



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from collections import namedtuple

from pyhocon import ConfigFactory, ConfigTree  # noqa: F401
from typing import Iterator, Union, Dict, Any  # noqa: F401

from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.sql_alchemy_extractor import SQLAlchemyExtractor
from databuilder.models.table_metadata import TableMetadata, ColumnMetadata
from itertools import groupby

TableKey = namedtuple('TableKey', ['schema_name', 'table_name'])

LOGGER = logging.getLogger(__name__)


class MSSQLMetadataExtractor(Extractor):
    """
    Extracts Microsoft SQL Server table and column metadata from underlying
    meta store database using SQLAlchemyExtractor
    """

    # SELECT statement from MS SQL to extract table and column metadata
    SQL_STATEMENT = """
        SELECT DISTINCT
            {cluster_source} AS cluster,
            TBL.TABLE_SCHEMA AS [schema_name],
            TBL.TABLE_NAME AS [name],
            CAST(PROP.VALUE AS NVARCHAR(MAX)) AS [description],
            COL.COLUMN_NAME AS [col_name],
            COL.DATA_TYPE AS [col_type],
            CAST(PROP_COL.VALUE AS NVARCHAR(MAX)) AS [col_description],
            COL.ORDINAL_POSITION AS col_sort_order
        FROM INFORMATION_SCHEMA.TABLES TBL
        INNER JOIN INFORMATION_SCHEMA.COLUMNS COL
            ON (COL.TABLE_NAME = TBL.TABLE_NAME
                AND COL.TABLE_SCHEMA = TBL.TABLE_SCHEMA )
        LEFT JOIN SYS.EXTENDED_PROPERTIES PROP
            ON (PROP.MAJOR_ID = OBJECT_ID(TBL.TABLE_SCHEMA + '.' + TBL.TABLE_NAME)
                AND PROP.MINOR_ID = 0
                AND PROP.NAME = 'MS_Description')
        LEFT JOIN SYS.EXTENDED_PROPERTIES PROP_COL
            ON (PROP_COL.MAJOR_ID = OBJECT_ID(TBL.TABLE_SCHEMA + '.' + TBL.TABLE_NAME)
                AND PROP_COL.MINOR_ID = COL.ORDINAL_POSITION
                AND PROP_COL.NAME = 'MS_Description')
        WHERE TBL.TABLE_TYPE = 'base table' {where_clause_suffix}
        ORDER BY
            CLUSTER,
            SCHEMA_NAME,
            NAME,
            COL_SORT_ORDER
        ;
    """

    # CONFIG KEYS
    WHERE_CLAUSE_SUFFIX_KEY = 'where_clause_suffix'
    CLUSTER_KEY = 'cluster_key'
    USE_CATALOG_AS_CLUSTER_NAME = 'use_catalog_as_cluster_name'
    DATABASE_KEY = 'database_key'

    # Default values
    DEFAULT_CLUSTER_NAME = 'DB_NAME()'

    DEFAULT_CONFIG = ConfigFactory.from_dict({
        WHERE_CLAUSE_SUFFIX_KEY: '',
        CLUSTER_KEY: DEFAULT_CLUSTER_NAME,
        USE_CATALOG_AS_CLUSTER_NAME: True}
    )

    DEFAULT_WHERE_CLAUSE_VALUE = 'and tbl.table_schema in {schemas}'

    def init(self, conf):
        # type: (ConfigTree) -> None
        conf = conf.with_fallback(MSSQLMetadataExtractor.DEFAULT_CONFIG)

        self._cluster = '{}'.format(
            conf.get_string(MSSQLMetadataExtractor.CLUSTER_KEY))

        if conf.get_bool(MSSQLMetadataExtractor.USE_CATALOG_AS_CLUSTER_NAME):
            cluster_source = "DB_NAME()"
        else:
            cluster_source = "'{}'".format(self._cluster)

        self._database = conf.get_string(
            MSSQLMetadataExtractor.DATABASE_KEY,
            default='mssql')

        config_where_clause = conf.get_string(
            MSSQLMetadataExtractor.WHERE_CLAUSE_SUFFIX_KEY)

        logging.info("Crawling for Schemas %s", config_where_clause)

        if len(config_where_clause) > 0:
            where_clause_suffix = MSSQLMetadataExtractor\
                .DEFAULT_WHERE_CLAUSE_VALUE\
                .format(schemas=config_where_clause)
        else:
            where_clause_suffix = ''

        self.sql_stmt = MSSQLMetadataExtractor.SQL_STATEMENT.format(
            where_clause_suffix=where_clause_suffix,
            cluster_source=cluster_source
        )

        LOGGER.info('SQL for MS SQL Metadata: {}'.format(self.sql_stmt))

        self._alchemy_extractor = SQLAlchemyExtractor()
        sql_alch_conf = Scoped\
            .get_scoped_conf(conf, self._alchemy_extractor.get_scope()) \
            .with_fallback(
                ConfigFactory.from_dict({
                    SQLAlchemyExtractor.EXTRACT_SQL: self.sql_stmt})
            )

        self._alchemy_extractor.init(sql_alch_conf)
        self._extract_iter = None  # type: Union[None, Iterator]

    def extract(self):
        # type: () -> Union[TableMetadata, None]
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def get_scope(self):
        # type: () -> str
        return 'extractor.mssql_metadata'

    def _get_extract_iter(self):
        # type: () -> Iterator[TableMetadata]
        """
        Using itertools.groupby and raw level iterator,
        it groups to table and yields TableMetadata
        :return:
        """
        for key, group in groupby(self._get_raw_extract_iter(), self._get_table_key):
            columns = []

            for row in group:
                last_row = row
                columns.append(
                    ColumnMetadata(
                        row['col_name'],
                        row['col_description'],
                        row['col_type'],
                        row['col_sort_order']))

            yield TableMetadata(
                self._database,
                last_row['cluster'],
                last_row['schema_name'],
                last_row['name'],
                last_row['description'],
                columns,
                tags=last_row['schema_name'])

    def _get_raw_extract_iter(self):
        # type: () -> Iterator[Dict[str, Any]]
        """
        Provides iterator of result row from SQLAlchemy extractor
        :return:
        """
        row = self._alchemy_extractor.extract()
        while row:
            yield row
            row = self._alchemy_extractor.extract()

    def _get_table_key(self, row):
        # type: (Dict[str, Any]) -> Union[TableKey, None]
        """
        Table key consists of schema and table name
        :param row:
        :return:
        """
        if row:
            return TableKey(
                schema_name=row['schema_name'],
                table_name=row['name'])

        return None



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import importlib
from typing import Iterable, Any  # noqa: F401

from pyhocon import ConfigTree  # noqa: F401

from databuilder.extractor.base_extractor import Extractor


class GenericExtractor(Extractor):
    """
    Extractor to extract any arbitrary values from users.
    """
    EXTRACTION_ITEMS = 'extraction_items'

    def init(self, conf):
        # type: (ConfigTree) -> None
        """
        Receives a list of dictionaries which is used for extraction
        :param conf:
        :return:
        """
        self.conf = conf
        self.values = conf.get(GenericExtractor.EXTRACTION_ITEMS)  # type: Iterable[Any]

        model_class = conf.get('model_class', None)
        if model_class:
            module_name, class_name = model_class.rsplit(".", 1)
            mod = importlib.import_module(module_name)
            self.model_class = getattr(mod, class_name)
            results = [self.model_class(**result)
                       for result in self.values]

            self._iter = iter(results)
        else:
            self._iter = iter(self.values)

    def extract(self):
        # type: () -> Any
        """
        Fetch one sql result row, convert to {model_class} if specified before
        returning.
        :return:
        """
        try:
            result = next(self._iter)
            return result
        except StopIteration:
            return None

    def get_scope(self):
        # type: () -> str
        return 'extractor.generic'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging

from pyhocon import ConfigTree  # noqa: F401
from typing import Any  # noqa: F401

from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.dashboard.mode_analytics.mode_dashboard_utils import ModeDashboardUtils
from databuilder.rest_api.mode_analytics.mode_paginated_rest_api_query import ModePaginatedRestApiQuery
from databuilder.rest_api.rest_api_query import RestApiQuery  # noqa: F401

LOGGER = logging.getLogger(__name__)


class ModeDashboardUsageExtractor(Extractor):
    """
    A Extractor that extracts Mode dashboard's accumulated view count
    """

    def init(self, conf):
        # type: (ConfigTree) -> None

        self._conf = conf

        restapi_query = self._build_restapi_query()
        self._extractor = ModeDashboardUtils.create_mode_rest_api_extractor(restapi_query=restapi_query,
                                                                            conf=self._conf)

    def extract(self):
        # type: () -> Any

        return self._extractor.extract()

    def get_scope(self):
        # type: () -> str

        return 'extractor.mode_dashboard_usage'

    def _build_restapi_query(self):
        """
        Build REST API Query. To get Mode Dashboard usage, it needs to call two APIs (spaces API and reports
        API) joining together.
        :return: A RestApiQuery that provides Mode Dashboard metadata
        """
        # type: () -> RestApiQuery

        # https://mode.com/developer/api-reference/analytics/reports/#listReportsInSpace
        reports_url_template = 'https://app.mode.com/api/{organization}/spaces/{dashboard_group_id}/reports'

        spaces_query = ModeDashboardUtils.get_spaces_query_api(conf=self._conf)
        params = ModeDashboardUtils.get_auth_params(conf=self._conf)

        # Reports
        # JSONPATH expression. it goes into array which is located in _embedded.reports and then extracts token,
        # and view_count
        json_path = '_embedded.reports[*].[token,view_count]'
        field_names = ['dashboard_id', 'accumulated_view_count']
        reports_query = ModePaginatedRestApiQuery(query_to_join=spaces_query, url=reports_url_template, params=params,
                                                  json_path=json_path, field_names=field_names, skip_no_result=True)
        return reports_query



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import json
import logging
from collections import namedtuple

import google.oauth2.service_account
import google_auth_httplib2
from googleapiclient.discovery import build
import httplib2
from pyhocon import ConfigTree  # noqa: F401
from typing import List, Any  # noqa: F401

from databuilder.extractor.base_extractor import Extractor


DatasetRef = namedtuple('DatasetRef', ['datasetId', 'projectId'])
TableKey = namedtuple('TableKey', ['schema', 'table_name'])

LOGGER = logging.getLogger(__name__)


class BaseBigQueryExtractor(Extractor):
    PROJECT_ID_KEY = 'project_id'
    KEY_PATH_KEY = 'key_path'
    # sometimes we don't have a key path, but only have an variable
    CRED_KEY = 'project_cred'
    PAGE_SIZE_KEY = 'page_size'
    FILTER_KEY = 'filter'
    _DEFAULT_SCOPES = ['https://www.googleapis.com/auth/bigquery.readonly', ]
    DEFAULT_PAGE_SIZE = 300
    NUM_RETRIES = 3
    DATE_LENGTH = 8

    def init(self, conf):
        # type: (ConfigTree) -> None
        # should use key_path, or cred_key if the former doesn't exist
        self.key_path = conf.get_string(BaseBigQueryExtractor.KEY_PATH_KEY, None)
        self.cred_key = conf.get_string(BaseBigQueryExtractor.CRED_KEY, None)
        self.project_id = conf.get_string(BaseBigQueryExtractor.PROJECT_ID_KEY)
        self.pagesize = conf.get_int(
            BaseBigQueryExtractor.PAGE_SIZE_KEY,
            BaseBigQueryExtractor.DEFAULT_PAGE_SIZE)
        self.filter = conf.get_string(BaseBigQueryExtractor.FILTER_KEY, '')

        if self.key_path:
            credentials = (
                google.oauth2.service_account.Credentials.from_service_account_file(
                    self.key_path, scopes=self._DEFAULT_SCOPES))
        else:
            if self.cred_key:
                service_account_info = json.loads(self.cred_key)
                credentials = (
                    google.oauth2.service_account.Credentials.from_service_account_info(
                        service_account_info, scopes=self._DEFAULT_SCOPES))
            else:
                credentials, _ = google.auth.default(scopes=self._DEFAULT_SCOPES)

        http = httplib2.Http()
        authed_http = google_auth_httplib2.AuthorizedHttp(credentials, http=http)
        self.bigquery_service = build('bigquery', 'v2', http=authed_http, cache_discovery=False)
        self.logging_service = build('logging', 'v2', http=authed_http, cache_discovery=False)
        self.iter = iter(self._iterate_over_tables())

    def extract(self):
        # type: () -> Any
        try:
            return next(self.iter)
        except StopIteration:
            return None

    def _is_sharded_table(self, table_id):
        suffix = table_id[-BaseBigQueryExtractor.DATE_LENGTH:]
        return suffix.isdigit()

    def _iterate_over_tables(self):
        # type: () -> Any
        for dataset in self._retrieve_datasets():
            for entry in self._retrieve_tables(dataset):
                yield(entry)

    def _retrieve_datasets(self):
        # type: () -> List[DatasetRef]
        datasets = []
        for page in self._page_dataset_list_results():
            if 'datasets' not in page:
                continue

            for dataset in page['datasets']:
                dataset_ref = dataset['datasetReference']
                ref = DatasetRef(**dataset_ref)
                datasets.append(ref)

        return datasets

    def _page_dataset_list_results(self):
        # type: () -> Any
        response = self.bigquery_service.datasets().list(
            projectId=self.project_id,
            all=False,  # Do not return hidden datasets
            filter=self.filter,
            maxResults=self.pagesize).execute(
                num_retries=BaseBigQueryExtractor.NUM_RETRIES)

        while response:
            yield response

            if 'nextPageToken' in response:
                response = self.bigquery_service.datasets().list(
                    projectId=self.project_id,
                    all=True,
                    filter=self.filter,
                    pageToken=response['nextPageToken']).execute(
                        num_retries=BaseBigQueryExtractor.NUM_RETRIES)
            else:
                response = None

    def _page_table_list_results(self, dataset):
        # type: (DatasetRef) -> Any
        response = self.bigquery_service.tables().list(
            projectId=dataset.projectId,
            datasetId=dataset.datasetId,
            maxResults=self.pagesize).execute(
                num_retries=BaseBigQueryExtractor.NUM_RETRIES)

        while response:
            yield response

            if 'nextPageToken' in response:
                response = self.bigquery_service.tables().list(
                    projectId=dataset.projectId,
                    datasetId=dataset.datasetId,
                    maxResults=self.pagesize,
                    pageToken=response['nextPageToken']).execute(
                        num_retries=BaseBigQueryExtractor.NUM_RETRIES)
            else:
                response = None

    def get_scope(self):
        # type: () -> str
        return 'extractor.bigquery_table_metadata'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging

from pyhocon import ConfigTree  # noqa: F401
from typing import Any, Dict  # noqa: F401

from databuilder.transformer.base_transformer import Transformer

TEMPLATE = 'template'
FIELD_NAME = 'field_name'  # field name to UPSERT

LOGGER = logging.getLogger(__name__)


class TemplateVariableSubstitutionTransformer(Transformer):
    """
    Add/Replace field in Dict by string.format based on given template and provide record Dict as a template parameter
    https://docs.python.org/3.4/library/string.html#string.Formatter.format

    """

    def init(self, conf):
        # type: (ConfigTree) -> None

        self._template = conf.get_string(TEMPLATE)
        self._field_name = conf.get_string(FIELD_NAME)

    def transform(self, record):
        # type: (Dict[str, Any]) -> Dict[str, Any]

        val = self._template.format(**record)
        record[self._field_name] = val
        return record

    def get_scope(self):
        # type: () -> str
        return 'transformer.template_variable_substitution'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import textwrap
from typing import Any  # noqa: F401

from pyhocon import ConfigTree  # noqa: F401

from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.neo4j_extractor import Neo4jExtractor
from databuilder.publisher.neo4j_csv_publisher import JOB_PUBLISH_TAG


class Neo4jSearchDataExtractor(Extractor):
    """
    Extractor to fetch data required to support search from Neo4j graph database
    Use Neo4jExtractor extractor class
    """
    CYPHER_QUERY_CONFIG_KEY = 'cypher_query'
    ENTITY_TYPE = 'entity_type'

    DEFAULT_NEO4J_TABLE_CYPHER_QUERY = textwrap.dedent(
        """
        MATCH (db:Database)<-[:CLUSTER_OF]-(cluster:Cluster)
        <-[:SCHEMA_OF]-(schema:Schema)<-[:TABLE_OF]-(table:Table)
        {publish_tag_filter}
        OPTIONAL MATCH (table)-[:DESCRIPTION]->(table_description:Description)
        OPTIONAL MATCH (schema)-[:DESCRIPTION]->(schema_description:Description)
        OPTIONAL MATCH (table)-[:DESCRIPTION]->(prog_descs:Programmatic_Description)
        WITH db, cluster, schema, schema_description, table, table_description,
        COLLECT(prog_descs.description) as programmatic_descriptions
        OPTIONAL MATCH (table)-[:TAGGED_BY]->(tags:Tag) WHERE tags.tag_type='default'
        WITH db, cluster, schema, schema_description, table, table_description, programmatic_descriptions,
        COLLECT(DISTINCT tags.key) as tags
        OPTIONAL MATCH (table)-[:TAGGED_BY]->(badges:Tag) WHERE badges.tag_type='badge'
        WITH db, cluster, schema, schema_description, table, table_description, programmatic_descriptions, tags,
        COLLECT(DISTINCT badges.key) as badges
        OPTIONAL MATCH (table)-[read:READ_BY]->(user:User)
        WITH db, cluster, schema, schema_description, table, table_description, programmatic_descriptions, tags, badges,
        SUM(read.read_count) AS total_usage,
        COUNT(DISTINCT user.email) as unique_usage
        OPTIONAL MATCH (table)-[:COLUMN]->(col:Column)
        OPTIONAL MATCH (col)-[:DESCRIPTION]->(col_description:Description)
        WITH db, cluster, schema, schema_description, table, table_description, tags, badges, total_usage, unique_usage,
        programmatic_descriptions,
        COLLECT(col.name) AS column_names, COLLECT(col_description.description) AS column_descriptions
        OPTIONAL MATCH (table)-[:LAST_UPDATED_AT]->(time_stamp:Timestamp)
        RETURN db.name as database, cluster.name AS cluster, schema.name AS schema,
        schema_description.description AS schema_description,
        table.name AS name, table.key AS key, table_description.description AS description,
        time_stamp.last_updated_timestamp AS last_updated_timestamp,
        column_names,
        column_descriptions,
        total_usage,
        unique_usage,
        tags,
        badges,
        programmatic_descriptions
        ORDER BY table.name;
        """
    )

    DEFAULT_NEO4J_USER_CYPHER_QUERY = textwrap.dedent(
        """
        MATCH (user:User)
        OPTIONAL MATCH (user)-[read:READ]->(a)
        OPTIONAL MATCH (user)-[own:OWNER_OF]->(b)
        OPTIONAL MATCH (user)-[follow:FOLLOWED_BY]->(c)
        OPTIONAL MATCH (user)-[manage_by:MANAGE_BY]->(manager)
        {publish_tag_filter}
        with user, a, b, c, read, own, follow, manager
        where user.full_name is not null
        return user.email as email, user.first_name as first_name, user.last_name as last_name,
        user.full_name as full_name, user.github_username as github_username, user.team_name as team_name,
        user.employee_type as employee_type, manager.email as manager_email,
        user.slack_id as slack_id, user.is_active as is_active, user.role_name as role_name,
        REDUCE(sum_r = 0, r in COLLECT(DISTINCT read)| sum_r + r.read_count) AS total_read,
        count(distinct b) as total_own,
        count(distinct c) AS total_follow
        order by user.email
        """
    )

    DEFAULT_NEO4J_DASHBOARD_CYPHER_QUERY = textwrap.dedent(
        """
        MATCH (dashboard:Dashboard)
        {publish_tag_filter}
        MATCH (dashboard)-[:DASHBOARD_OF]->(dbg:Dashboardgroup)
        MATCH (dbg)-[:DASHBOARD_GROUP_OF]->(cluster:Cluster)
        OPTIONAL MATCH (dashboard)-[:DESCRIPTION]->(db_descr:Description)
        OPTIONAL MATCH (dbg)-[:DESCRIPTION]->(dbg_descr:Description)
        OPTIONAL MATCH (dashboard)-[:EXECUTED]->(last_exec:Execution)
        WHERE split(last_exec.key, '/')[5] = '_last_successful_execution'
        OPTIONAL MATCH (dashboard)-[read:READ_BY]->(user:User)
        WITH dashboard, dbg, db_descr, dbg_descr, cluster, last_exec, SUM(read.read_count) AS total_usage
        OPTIONAL MATCH (dashboard)-[:HAS_QUERY]->(query:Query)
        WITH dashboard, dbg, db_descr, dbg_descr, cluster, last_exec, COLLECT(DISTINCT query.name) as query_names,
        total_usage
        OPTIONAL MATCH (dashboard)-[:TAGGED_BY]->(tags:Tag) WHERE tags.tag_type='default'
        WITH dashboard, dbg, db_descr, dbg_descr, cluster, last_exec, query_names, total_usage,
        COLLECT(DISTINCT tags.key) as tags
        OPTIONAL MATCH (dashboard)-[:TAGGED_BY]->(badges:Tag) WHERE badges.tag_type='badge'
        WITH  dashboard, dbg, db_descr, dbg_descr, cluster, last_exec, query_names, total_usage, tags,
        COLLECT(DISTINCT badges.key) as badges
        RETURN dbg.name as group_name, dashboard.name as name, cluster.name as cluster,
        coalesce(db_descr.description, '') as description,
        coalesce(dbg.description, '') as group_description, dbg.dashboard_group_url as group_url,
        dashboard.dashboard_url as url, dashboard.key as uri,
        split(dashboard.key, '_')[0] as product, toInteger(last_exec.timestamp) as last_successful_run_timestamp,
        query_names, total_usage, tags, badges
        order by dbg.name;
        """
    )

    # todo: we will add more once we add more entities
    DEFAULT_QUERY_BY_ENTITY = {
        'table': DEFAULT_NEO4J_TABLE_CYPHER_QUERY,
        'user': DEFAULT_NEO4J_USER_CYPHER_QUERY,
        'dashboard': DEFAULT_NEO4J_DASHBOARD_CYPHER_QUERY
    }

    def init(self, conf):
        # type: (ConfigTree) -> None
        """
        Initialize Neo4jExtractor object from configuration and use that for extraction
        """
        self.conf = conf
        self.entity = conf.get_string(Neo4jSearchDataExtractor.ENTITY_TYPE, default='table').lower()
        # extract cypher query from conf, if specified, else use default query
        if Neo4jSearchDataExtractor.CYPHER_QUERY_CONFIG_KEY in conf:
            self.cypher_query = conf.get_string(Neo4jSearchDataExtractor.CYPHER_QUERY_CONFIG_KEY)
        else:
            default_query = Neo4jSearchDataExtractor.DEFAULT_QUERY_BY_ENTITY[self.entity]
            self.cypher_query = self._add_publish_tag_filter(conf.get_string(JOB_PUBLISH_TAG, ''),
                                                             cypher_query=default_query)

        self.neo4j_extractor = Neo4jExtractor()
        # write the cypher query in configs in Neo4jExtractor scope
        key = self.neo4j_extractor.get_scope() + '.' + Neo4jExtractor.CYPHER_QUERY_CONFIG_KEY
        self.conf.put(key, self.cypher_query)
        # initialize neo4j_extractor from configs
        self.neo4j_extractor.init(Scoped.get_scoped_conf(self.conf, self.neo4j_extractor.get_scope()))

    def close(self):
        # type: () -> None
        """
        Use close() method specified by neo4j_extractor
        to close connection to neo4j cluster
        """
        self.neo4j_extractor.close()

    def extract(self):
        # type: () -> Any
        """
        Invoke extract() method defined by neo4j_extractor
        """
        return self.neo4j_extractor.extract()

    def get_scope(self):
        # type: () -> str
        return 'extractor.search_data'

    def _add_publish_tag_filter(self, publish_tag, cypher_query):
        """
        Adds publish tag filter into Cypher query
        :param publish_tag: value of publish tag.
        :param cypher_query:
        :return:
        """
        # type: (str, str) -> str
        if not publish_tag:
            publish_tag_filter = ''
        else:
            if not hasattr(self, 'entity'):
                self.entity = 'table'
            publish_tag_filter = """WHERE {entity}.published_tag = '{tag}'""".format(entity=self.entity,
                                                                                     tag=publish_tag)
        return cypher_query.format(publish_tag_filter=publish_tag_filter)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from collections import namedtuple

from typing import Any, Union, Iterator, Dict, Set, Optional  # noqa: F401

from databuilder.models.cluster import cluster_constants
from databuilder.models.neo4j_csv_serde import (
    Neo4jCsvSerializable, NODE_LABEL, NODE_KEY, RELATION_START_KEY, RELATION_END_KEY, RELATION_START_LABEL,
    RELATION_END_LABEL, RELATION_TYPE, RELATION_REVERSE_TYPE)
# TODO: We could separate TagMetadata from table_metadata to own module
from databuilder.models.table_metadata import TagMetadata

NodeTuple = namedtuple('KeyName', ['key', 'name', 'label'])
RelTuple = namedtuple('RelKeys', ['start_label', 'end_label', 'start_key', 'end_key', 'type', 'reverse_type'])


class DashboardMetadata(Neo4jCsvSerializable):
    """
    Dashboard metadata that contains dashboard group name, dashboardgroup description, dashboard description,
    along with tags, owner userid and lastreloadtime.
    (Owner ID and last reload time will be supported by separate Extractor later on with more information)

    It implements Neo4jCsvSerializable so that it can be serialized to produce
    Dashboard, Tag, Description, Lastreloadtime and relation of those. Additionally, it will create
    Dashboardgroup with relationships to Dashboard. If users exist in neo4j, it will create
    the relation between dashboard and user (owner).

    Lastreloadtime is the time when the Dashboard was last reloaded.
    """
    CLUSTER_KEY_FORMAT = '{product}_dashboard://{cluster}'
    CLUSTER_DASHBOARD_GROUP_RELATION_TYPE = 'DASHBOARD_GROUP'
    DASHBOARD_GROUP_CLUSTER_RELATION_TYPE = 'DASHBOARD_GROUP_OF'

    DASHBOARD_NODE_LABEL = 'Dashboard'
    DASHBOARD_KEY_FORMAT = '{product}_dashboard://{cluster}.{dashboard_group}/{dashboard_name}'
    DASHBOARD_NAME = 'name'
    DASHBOARD_CREATED_TIME_STAMP = 'created_timestamp'
    DASHBOARD_GROUP_URL = 'dashboard_group_url'
    DASHBOARD_URL = 'dashboard_url'

    DASHBOARD_DESCRIPTION_NODE_LABEL = 'Description'
    DASHBOARD_DESCRIPTION = 'description'
    DASHBOARD_DESCRIPTION_FORMAT = \
        '{product}_dashboard://{cluster}.{dashboard_group}/{dashboard_name}/_description'
    DASHBOARD_DESCRIPTION_RELATION_TYPE = 'DESCRIPTION'
    DESCRIPTION_DASHBOARD_RELATION_TYPE = 'DESCRIPTION_OF'

    DASHBOARD_GROUP_NODE_LABEL = 'Dashboardgroup'
    DASHBOARD_GROUP_KEY_FORMAT = '{product}_dashboard://{cluster}.{dashboard_group}'
    DASHBOARD_GROUP_DASHBOARD_RELATION_TYPE = 'DASHBOARD'
    DASHBOARD_DASHBOARD_GROUP_RELATION_TYPE = 'DASHBOARD_OF'

    DASHBOARD_GROUP_DESCRIPTION_KEY_FORMAT = '{product}_dashboard://{cluster}.{dashboard_group}/_description'

    DASHBOARD_TAG_RELATION_TYPE = 'TAG'
    TAG_DASHBOARD_RELATION_TYPE = 'TAG_OF'

    serialized_nodes = set()  # type: Set[Any]
    serialized_rels = set()  # type: Set[Any]

    def __init__(self,
                 dashboard_group,  # type: str
                 dashboard_name,  # type: str
                 description,  # type: Union[str, None]
                 tags=None,  # type: List
                 cluster='gold',  # type: str
                 product='',  # type: Optional[str]
                 dashboard_group_id=None,  # type: Optional[str]
                 dashboard_id=None,  # type: Optional[str]
                 dashboard_group_description=None,  # type: Optional[str]
                 created_timestamp=None,  # type: Optional[int]
                 dashboard_group_url=None,  # type: Optional[str]
                 dashboard_url=None,  # type: Optional[str]
                 **kwargs
                 ):
        # type: (...) -> None

        self.dashboard_group = dashboard_group
        self.dashboard_name = dashboard_name
        self.dashboard_group_id = dashboard_group_id if dashboard_group_id else dashboard_group
        self.dashboard_id = dashboard_id if dashboard_id else dashboard_name
        self.description = description
        self.tags = tags
        self.product = product
        self.cluster = cluster
        self.dashboard_group_description = dashboard_group_description
        self.created_timestamp = created_timestamp
        self.dashboard_group_url = dashboard_group_url
        self.dashboard_url = dashboard_url
        self._processed_cluster = set()
        self._processed_dashboard_group = set()
        self._node_iterator = self._create_next_node()
        self._relation_iterator = self._create_next_relation()

    def __repr__(self):
        # type: () -> str
        return 'DashboardMetadata({!r}, {!r}, {!r}, {!r}, {!r}, {!r}, {!r}, {!r}, {!r}, {!r})' \
            .format(self.dashboard_group,
                    self.dashboard_name,
                    self.description,
                    self.tags,
                    self.dashboard_group_id,
                    self.dashboard_id,
                    self.dashboard_group_description,
                    self.created_timestamp,
                    self.dashboard_group_url,
                    self.dashboard_url,
                    )

    def _get_cluster_key(self):
        # type: () -> str
        return DashboardMetadata.CLUSTER_KEY_FORMAT.format(cluster=self.cluster,
                                                           product=self.product)

    def _get_dashboard_key(self):
        # type: () -> str
        return DashboardMetadata.DASHBOARD_KEY_FORMAT.format(dashboard_group=self.dashboard_group_id,
                                                             dashboard_name=self.dashboard_id,
                                                             cluster=self.cluster,
                                                             product=self.product)

    def _get_dashboard_description_key(self):
        # type: () -> str
        return DashboardMetadata.DASHBOARD_DESCRIPTION_FORMAT.format(dashboard_group=self.dashboard_group_id,
                                                                     dashboard_name=self.dashboard_id,
                                                                     cluster=self.cluster,
                                                                     product=self.product)

    def _get_dashboard_group_description_key(self):
        # type: () -> str
        return DashboardMetadata.DASHBOARD_GROUP_DESCRIPTION_KEY_FORMAT.format(dashboard_group=self.dashboard_group_id,
                                                                               cluster=self.cluster,
                                                                               product=self.product)

    def _get_dashboard_group_key(self):
        # type: () -> str
        return DashboardMetadata.DASHBOARD_GROUP_KEY_FORMAT.format(dashboard_group=self.dashboard_group_id,
                                                                   cluster=self.cluster,
                                                                   product=self.product)

    def _get_dashboard_last_reload_time_key(self):
        # type: () -> str
        return DashboardMetadata.DASHBOARD_LAST_RELOAD_TIME_FORMAT.format(dashboard_group=self.dashboard_group,
                                                                          dashboard_name=self.dashboard_id,
                                                                          cluster=self.cluster,
                                                                          product=self.product)

    def create_next_node(self):
        # type: () -> Union[Dict[str, Any], None]
        try:
            return next(self._node_iterator)
        except StopIteration:
            return None

    def _create_next_node(self):
        # type: () -> Iterator[Any]
        # Cluster node
        if not self._get_cluster_key() in self._processed_cluster:
            self._processed_cluster.add(self._get_cluster_key())
            yield {
                NODE_LABEL: cluster_constants.CLUSTER_NODE_LABEL,
                NODE_KEY: self._get_cluster_key(),
                cluster_constants.CLUSTER_NAME_PROP_KEY: self.cluster
            }

        # Dashboard node
        dashboard_node = {
            NODE_LABEL: DashboardMetadata.DASHBOARD_NODE_LABEL,
            NODE_KEY: self._get_dashboard_key(),
            DashboardMetadata.DASHBOARD_NAME: self.dashboard_name,
        }
        if self.created_timestamp:
            dashboard_node[DashboardMetadata.DASHBOARD_CREATED_TIME_STAMP] = self.created_timestamp

        if self.dashboard_url:
            dashboard_node[DashboardMetadata.DASHBOARD_URL] = self.dashboard_url

        yield dashboard_node

        # Dashboard group
        if self.dashboard_group and not self._get_dashboard_group_key() in self._processed_dashboard_group:
            self._processed_dashboard_group.add(self._get_dashboard_group_key())
            dashboard_group_node = {
                NODE_LABEL: DashboardMetadata.DASHBOARD_GROUP_NODE_LABEL,
                NODE_KEY: self._get_dashboard_group_key(),
                DashboardMetadata.DASHBOARD_NAME: self.dashboard_group,
            }

            if self.dashboard_group_url:
                dashboard_group_node[DashboardMetadata.DASHBOARD_GROUP_URL] = self.dashboard_group_url

            yield dashboard_group_node

        # Dashboard group description
        if self.dashboard_group_description:
            yield {NODE_LABEL: DashboardMetadata.DASHBOARD_DESCRIPTION_NODE_LABEL,
                   NODE_KEY: self._get_dashboard_group_description_key(),
                   DashboardMetadata.DASHBOARD_DESCRIPTION: self.dashboard_group_description}

        # Dashboard description node
        if self.description:
            yield {NODE_LABEL: DashboardMetadata.DASHBOARD_DESCRIPTION_NODE_LABEL,
                   NODE_KEY: self._get_dashboard_description_key(),
                   DashboardMetadata.DASHBOARD_DESCRIPTION: self.description}

        # Dashboard tag node
        if self.tags:
            for tag in self.tags:
                yield {NODE_LABEL: TagMetadata.TAG_NODE_LABEL,
                       NODE_KEY: TagMetadata.get_tag_key(tag),
                       TagMetadata.TAG_TYPE: 'dashboard'}

    def create_next_relation(self):
        # type: () -> Union[Dict[str, Any], None]
        try:
            return next(self._relation_iterator)
        except StopIteration:
            return None

    def _create_next_relation(self):
        # type: () -> Iterator[Any]

        # Cluster <-> Dashboard group
        yield {
            RELATION_START_LABEL: cluster_constants.CLUSTER_NODE_LABEL,
            RELATION_END_LABEL: DashboardMetadata.DASHBOARD_GROUP_NODE_LABEL,
            RELATION_START_KEY: self._get_cluster_key(),
            RELATION_END_KEY: self._get_dashboard_group_key(),
            RELATION_TYPE: DashboardMetadata.CLUSTER_DASHBOARD_GROUP_RELATION_TYPE,
            RELATION_REVERSE_TYPE: DashboardMetadata.DASHBOARD_GROUP_CLUSTER_RELATION_TYPE
        }

        # Dashboard group > Dashboard group description relation
        if self.dashboard_group_description:
            yield {
                RELATION_START_LABEL: DashboardMetadata.DASHBOARD_GROUP_NODE_LABEL,
                RELATION_END_LABEL: DashboardMetadata.DASHBOARD_DESCRIPTION_NODE_LABEL,
                RELATION_START_KEY: self._get_dashboard_group_key(),
                RELATION_END_KEY: self._get_dashboard_group_description_key(),
                RELATION_TYPE: DashboardMetadata.DASHBOARD_DESCRIPTION_RELATION_TYPE,
                RELATION_REVERSE_TYPE: DashboardMetadata.DESCRIPTION_DASHBOARD_RELATION_TYPE
            }

        # Dashboard group > Dashboard relation
        yield {
            RELATION_START_LABEL: DashboardMetadata.DASHBOARD_NODE_LABEL,
            RELATION_END_LABEL: DashboardMetadata.DASHBOARD_GROUP_NODE_LABEL,
            RELATION_START_KEY: self._get_dashboard_key(),
            RELATION_END_KEY: self._get_dashboard_group_key(),
            RELATION_TYPE: DashboardMetadata.DASHBOARD_DASHBOARD_GROUP_RELATION_TYPE,
            RELATION_REVERSE_TYPE: DashboardMetadata.DASHBOARD_GROUP_DASHBOARD_RELATION_TYPE
        }

        # Dashboard > Dashboard description relation
        if self.description:
            yield {
                RELATION_START_LABEL: DashboardMetadata.DASHBOARD_NODE_LABEL,
                RELATION_END_LABEL: DashboardMetadata.DASHBOARD_DESCRIPTION_NODE_LABEL,
                RELATION_START_KEY: self._get_dashboard_key(),
                RELATION_END_KEY: self._get_dashboard_description_key(),
                RELATION_TYPE: DashboardMetadata.DASHBOARD_DESCRIPTION_RELATION_TYPE,
                RELATION_REVERSE_TYPE: DashboardMetadata.DESCRIPTION_DASHBOARD_RELATION_TYPE
            }

        # Dashboard > Dashboard tag relation
        if self.tags:
            for tag in self.tags:
                yield {
                    RELATION_START_LABEL: DashboardMetadata.DASHBOARD_NODE_LABEL,
                    RELATION_END_LABEL: TagMetadata.TAG_NODE_LABEL,
                    RELATION_START_KEY: self._get_dashboard_key(),
                    RELATION_END_KEY: TagMetadata.get_tag_key(tag),
                    RELATION_TYPE: DashboardMetadata.DASHBOARD_TAG_RELATION_TYPE,
                    RELATION_REVERSE_TYPE: DashboardMetadata.TAG_DASHBOARD_RELATION_TYPE
                }



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from collections import namedtuple

from pyhocon import ConfigFactory, ConfigTree  # noqa: F401
from typing import Iterator, Union, Dict, Any  # noqa: F401

from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.sql_alchemy_extractor import SQLAlchemyExtractor
from databuilder.models.table_metadata import TableMetadata, ColumnMetadata
from itertools import groupby


TableKey = namedtuple('TableKey', ['schema', 'table_name'])

LOGGER = logging.getLogger(__name__)


class PostgresMetadataExtractor(Extractor):
    """
    Extracts Postgres table and column metadata from underlying meta store database using SQLAlchemyExtractor
    """
    # SELECT statement from postgres information_schema to extract table and column metadata
    SQL_STATEMENT = """
    SELECT
      {cluster_source} as cluster, c.table_schema as schema, c.table_name as name, pgtd.description as description
      ,c.column_name as col_name, c.data_type as col_type
      , pgcd.description as col_description, ordinal_position as col_sort_order
    FROM INFORMATION_SCHEMA.COLUMNS c
    INNER JOIN
      pg_catalog.pg_statio_all_tables as st on c.table_schema=st.schemaname and c.table_name=st.relname
    LEFT JOIN
      pg_catalog.pg_description pgcd on pgcd.objoid=st.relid and pgcd.objsubid=c.ordinal_position
    LEFT JOIN
      pg_catalog.pg_description pgtd on pgtd.objoid=st.relid and pgtd.objsubid=0
    {where_clause_suffix}
    ORDER by cluster, schema, name, col_sort_order ;
    """

    # CONFIG KEYS
    WHERE_CLAUSE_SUFFIX_KEY = 'where_clause_suffix'
    CLUSTER_KEY = 'cluster_key'
    USE_CATALOG_AS_CLUSTER_NAME = 'use_catalog_as_cluster_name'
    DATABASE_KEY = 'database_key'

    # Default values
    DEFAULT_CLUSTER_NAME = 'master'

    DEFAULT_CONFIG = ConfigFactory.from_dict(
        {WHERE_CLAUSE_SUFFIX_KEY: ' ', CLUSTER_KEY: DEFAULT_CLUSTER_NAME, USE_CATALOG_AS_CLUSTER_NAME: True}
    )

    def init(self, conf):
        # type: (ConfigTree) -> None
        conf = conf.with_fallback(PostgresMetadataExtractor.DEFAULT_CONFIG)
        self._cluster = '{}'.format(conf.get_string(PostgresMetadataExtractor.CLUSTER_KEY))

        if conf.get_bool(PostgresMetadataExtractor.USE_CATALOG_AS_CLUSTER_NAME):
            cluster_source = "c.table_catalog"
        else:
            cluster_source = "'{}'".format(self._cluster)

        self._database = conf.get_string(PostgresMetadataExtractor.DATABASE_KEY, default='postgres')

        self.sql_stmt = PostgresMetadataExtractor.SQL_STATEMENT.format(
            where_clause_suffix=conf.get_string(PostgresMetadataExtractor.WHERE_CLAUSE_SUFFIX_KEY),
            cluster_source=cluster_source
        )

        self._alchemy_extractor = SQLAlchemyExtractor()
        sql_alch_conf = Scoped.get_scoped_conf(conf, self._alchemy_extractor.get_scope())\
            .with_fallback(ConfigFactory.from_dict({SQLAlchemyExtractor.EXTRACT_SQL: self.sql_stmt}))

        self.sql_stmt = sql_alch_conf.get_string(SQLAlchemyExtractor.EXTRACT_SQL)

        LOGGER.info('SQL for postgres metadata: {}'.format(self.sql_stmt))

        self._alchemy_extractor.init(sql_alch_conf)
        self._extract_iter = None  # type: Union[None, Iterator]

    def extract(self):
        # type: () -> Union[TableMetadata, None]
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def get_scope(self):
        # type: () -> str
        return 'extractor.postgres_metadata'

    def _get_extract_iter(self):
        # type: () -> Iterator[TableMetadata]
        """
        Using itertools.groupby and raw level iterator, it groups to table and yields TableMetadata
        :return:
        """
        for key, group in groupby(self._get_raw_extract_iter(), self._get_table_key):
            columns = []

            for row in group:
                last_row = row
                columns.append(ColumnMetadata(row['col_name'], row['col_description'],
                                              row['col_type'], row['col_sort_order']))

            yield TableMetadata(self._database, last_row['cluster'],
                                last_row['schema'],
                                last_row['name'],
                                last_row['description'],
                                columns)

    def _get_raw_extract_iter(self):
        # type: () -> Iterator[Dict[str, Any]]
        """
        Provides iterator of result row from SQLAlchemy extractor
        :return:
        """
        row = self._alchemy_extractor.extract()
        while row:
            yield row
            row = self._alchemy_extractor.extract()

    def _get_table_key(self, row):
        # type: (Dict[str, Any]) -> Union[TableKey, None]
        """
        Table key consists of schema and table name
        :param row:
        :return:
        """
        if row:
            return TableKey(schema=row['schema'], table_name=row['name'])

        return None



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import importlib
import logging

from pyhocon import ConfigTree  # noqa: F401
from typing import Any, Dict  # noqa: F401

from databuilder.transformer.base_transformer import Transformer

MODEL_CLASS = 'model_class'

LOGGER = logging.getLogger(__name__)


class DictToModel(Transformer):
    """
    Transforms dictionary into model
    """

    def init(self, conf):
        # type: (ConfigTree) -> None

        model_class = conf.get_string(MODEL_CLASS)
        module_name, class_name = model_class.rsplit(".", 1)
        mod = importlib.import_module(module_name)
        self._model_class = getattr(mod, class_name)

    def transform(self, record):
        # type: (Dict[str, Any]) -> Dict[str, Any]

        return self._model_class(**record)

    def get_scope(self):
        # type: () -> str
        return 'transformer.dict_to_model'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import csv
import importlib
from collections import defaultdict

from pyhocon import ConfigTree  # noqa: F401
from typing import Any, Iterator  # noqa: F401

from databuilder.extractor.base_extractor import Extractor
from databuilder.models.table_metadata import TableMetadata, ColumnMetadata


class CsvExtractor(Extractor):
    # Config keys
    FILE_LOCATION = 'file_location'

    """
    An Extractor that extracts records via CSV.
    """

    def init(self, conf):
        # type: (ConfigTree) -> None
        """
        :param conf:
        """
        self.conf = conf
        self.file_location = conf.get_string(CsvExtractor.FILE_LOCATION)

        model_class = conf.get('model_class', None)
        if model_class:
            module_name, class_name = model_class.rsplit(".", 1)
            mod = importlib.import_module(module_name)
            self.model_class = getattr(mod, class_name)
        self._load_csv()

    def _load_csv(self):
        # type: () -> None
        """
        Create an iterator to execute sql.
        """
        if not hasattr(self, 'results'):
            with open(self.file_location, 'r') as fin:
                self.results = [dict(i) for i in csv.DictReader(fin)]

        if hasattr(self, 'model_class'):
            results = [self.model_class(**result)
                       for result in self.results]
        else:
            results = self.results
        self.iter = iter(results)

    def extract(self):
        # type: () -> Any
        """
        Yield the csv result one at a time.
        convert the result to model if a model_class is provided
        """
        try:
            return next(self.iter)
        except StopIteration:
            return None
        except Exception as e:
            raise e

    def get_scope(self):
        # type: () -> str
        return 'extractor.csv'


class CsvTableColumnExtractor(Extractor):
    # Config keys
    TABLE_FILE_LOCATION = 'table_file_location'
    COLUMN_FILE_LOCATION = 'column_file_location'

    """
    An Extractor that combines Table and Column CSVs.
    """

    def init(self, conf):
        # type: (ConfigTree) -> None
        """
        :param conf:
        """
        self.conf = conf
        self.table_file_location = conf.get_string(CsvTableColumnExtractor.TABLE_FILE_LOCATION)
        self.column_file_location = conf.get_string(CsvTableColumnExtractor.COLUMN_FILE_LOCATION)
        self._load_csv()

    def _get_key(self, db, cluster, schema, tbl):
        return TableMetadata.TABLE_KEY_FORMAT.format(db=db,
                                                     cluster=cluster,
                                                     schema=schema,
                                                     tbl=tbl)

    def _load_csv(self):
        # type: () -> None
        """
        Create an iterator to execute sql.
        """

        with open(self.column_file_location, 'r') as fin:
            self.columns = [dict(i) for i in csv.DictReader(fin)]

        parsed_columns = defaultdict(list)
        for column_dict in self.columns:
            db = column_dict['database']
            cluster = column_dict['cluster']
            schema = column_dict['schema']
            table = column_dict['table_name']
            id = self._get_key(db, cluster, schema, table)
            column = ColumnMetadata(
                name=column_dict['name'],
                description=column_dict['description'],
                col_type=column_dict['col_type'],
                sort_order=int(column_dict['sort_order'])
            )
            parsed_columns[id].append(column)

        # Create Table Dictionary
        with open(self.table_file_location, 'r') as fin:
            tables = [dict(i) for i in csv.DictReader(fin)]

        results = []
        for table_dict in tables:
            db = table_dict['database']
            cluster = table_dict['cluster']
            schema = table_dict['schema']
            table = table_dict['name']
            id = self._get_key(db, cluster, schema, table)
            columns = parsed_columns[id]
            if columns is None:
                columns = []
            table = TableMetadata(database=table_dict['database'],
                                  cluster=table_dict['cluster'],
                                  schema=table_dict['schema'],
                                  name=table_dict['name'],
                                  description=table_dict['description'],
                                  columns=columns,
                                  is_view=table_dict['is_view'],
                                  tags=table_dict['tags']
                                  )
            results.append(table)
        self._iter = iter(results)

    def extract(self):
        # type: () -> Any
        """
        Yield the csv result one at a time.
        convert the result to model if a model_class is provided
        """
        try:
            return next(self._iter)
        except StopIteration:
            return None
        except Exception as e:
            raise e

    def get_scope(self):
        # type: () -> str
        return 'extractor.csvtablecolumn'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import abc

from requests.exceptions import HTTPError
from typing import Iterable, Union, List, Dict, Any, Optional  # noqa: F401


class BaseFailureHandler(object, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def can_skip_failure(self,
                         exception,  # type: Exception
                         ):
        # type: (...) -> bool
        pass


class HttpFailureSkipOnStatus(BaseFailureHandler):

    def __init__(self,
                 status_codes_to_skip,  # type: Iterable[int]
                 ):
        # type: (...) -> None
        self._status_codes_to_skip = {v for v in status_codes_to_skip}

    def can_skip_failure(self,
                         exception,  # type: Exception
                         ):
        # type: (...) -> bool

        if (isinstance(exception, HTTPError) or hasattr(exception, 'response')) \
                and exception.response.status_code in self._status_codes_to_skip:
            return True

        return False



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from pyhocon import ConfigFactory

from databuilder.transformer.base_transformer import Transformer
from databuilder.models.table_metadata import TableMetadata


class TableTagTransformer(Transformer):
    """Simple transformer that adds tags to all table nodes produced as part of a job."""
    # Config
    TAGS = 'tags'
    DEFAULT_CONFIG = ConfigFactory.from_dict({TAGS: None})

    def init(self, conf):
        conf = conf.with_fallback(TableTagTransformer.DEFAULT_CONFIG)
        tags = conf.get_string(TableTagTransformer.TAGS)

        self.tags = TableMetadata.format_tags(tags)

    def transform(self, record):
        if isinstance(record, TableMetadata):
            if record.tags:
                record.tags += self.tags
            else:
                record.tags = self.tags
        return record

    def get_scope(self):
        # type: () -> str
        return 'transformer.table_tag'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from databuilder.rest_api.rest_api_query import RestApiQuery


def sort_widgets(widgets):
    # type: (Iterator[Dict[str, Any]]) -> Iterator[Dict[str, Any]]
    """
    Sort raw widget data (as returned from the API) according to the position
    of the widgets in the dashboard (top to bottom, left to right)
    Redash does not return widgets in order of their position,
    so we do this to ensure that we look at widgets in a sensible order.
    """

    def row_and_col(widget):
        # these entities usually but not always have explicit rows and cols
        pos = widget['options'].get('position', {})
        return (pos.get('row', 0), pos.get('col', 0))

    return sorted(widgets, key=row_and_col)


def get_text_widgets(widgets):
    # type: (Iterator[Dict[str, Any]]) -> List[RedashTextWidget]
    """
    From the raw set of widget data returned from the API, filter down
    to text widgets and return them as a list of `RedashTextWidget`
    """

    return [RedashTextWidget(widget) for widget in widgets
            if 'text' in widget and 'visualization' not in widget]


def get_visualization_widgets(widgets):
    # type: (Iterator[Dict[str, Any]]) -> List[RedashVisualizationWidget]
    """
    From the raw set of widget data returned from the API, filter down
    to visualization widgets and return them as a list of `RedashVisualizationWidget`
    """

    return [RedashVisualizationWidget(widget) for widget in widgets
            if 'visualization' in widget]


def get_auth_headers(api_key):
    # type: (str) -> Dict[str, str]
    return {'Authorization': 'Key {}'.format(api_key)}


def generate_dashboard_description(text_widgets, viz_widgets):
    # type: (Iterator[RedashTextWidget], Iterator[RedashVisualizationWidget]) -> str
    """
    Redash doesn't have dashboard descriptions, so we'll make our own.
    If there exist any text widgets, concatenate them,
    and use this text as the description for this dashboard.
    If not, put together a list of query names.
    If all else fails, this looks like an empty dashboard.
    """

    if len(text_widgets) > 0:
        return '\n\n'.join([w.text for w in text_widgets])
    elif len(viz_widgets) > 0:
        query_list = '\n'.join(['- {}'.format(v.query_name) for v in set(viz_widgets)])
        return 'A dashboard containing the following queries:\n\n' + query_list

    return 'This dashboard appears to be empty!'


class RedashVisualizationWidget:
    """
    A visualization widget in a Redash dashboard.
    These are mapped 1:1 with queries, and can be of various types, e.g.:
    CHART, TABLE, PIVOT, etc.
    The query name acts like a title for the widget on the dashboard.
    """

    def __init__(self, data):
        # type: (Dict[str, Any]) -> None
        self._data = data

    @property
    def raw_query(self):
        # type () -> str
        return self._data['visualization']['query']['query']

    @property
    def data_source_id(self):
        # type: () -> int
        return self._data['visualization']['query']['data_source_id']

    @property
    def query_id(self):
        # type: () -> int
        return self._data['visualization']['query']['id']

    @property
    def query_relative_url(self):
        # type: () -> str
        return '/queries/{id}'.format(id=self.query_id)

    @property
    def query_name(self):
        # type: () -> str
        return self._data['visualization']['query']['name']


class RedashTextWidget:
    """
    A textbox in a Redash dashboad.
    It pretty much just contains a single text property (Markdown).
    """

    def __init__(self, data):
        # type: (Dict[str, Any]) -> None
        self._data = data

    @property
    def text(self):
        # type: () -> str
        return self._data['text']


class RedashPaginatedRestApiQuery(RestApiQuery):
    """
    Paginated Redash API queries
    """

    def __init__(self, **kwargs):
        # type: (...) -> None
        super(RedashPaginatedRestApiQuery, self).__init__(**kwargs)
        if 'params' not in self._params:
            self._params['params'] = {}
        self._params['params']['page'] = 1

    def _total_records(self, res):
        # type: (Dict[str, Any]) -> int
        return res['count']

    def _max_record_on_page(self, res):
        # type: (Dict[str, Any]) -> int
        return res['page_size'] * res['page']

    def _next_page(self, res):
        # type: (Dict[str, Any]) -> int
        return res['page'] + 1

    def _post_process(self, response):
        # type: (Any) -> None
        parsed = response.json()

        if self._max_record_on_page(parsed) >= self._total_records(parsed):
            self._more_pages = False
        else:
            self._params['params']['page'] = self._next_page(parsed)
            self._more_pages = True



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import List, Optional  # noqa: F401

from databuilder.models.elasticsearch_document import ElasticsearchDocument


class TableESDocument(ElasticsearchDocument):
    """
    Schema for the Search index document
    """

    def __init__(self,
                 database,  # type: str
                 cluster,  # type: str
                 schema,  # type: str
                 name,  # type: str
                 key,  # type: str
                 description,  # type: str
                 last_updated_timestamp,  # type: Optional[int]
                 column_names,  # type: List[str]
                 column_descriptions,  # type: List[str]
                 total_usage,  # type: int
                 unique_usage,  # type: int
                 tags,  # type: List[str],
                 badges=None,  # type: Optional[List[str]]
                 display_name=None,  # type: Optional[str]
                 schema_description=None,  # type: Optional[str]
                 programmatic_descriptions=[],  # type: List[str]
                 ):
        # type: (...) -> None
        self.database = database
        self.cluster = cluster
        self.schema = schema
        self.name = name
        self.display_name = display_name if display_name else '{schema}.{table}'.format(schema=schema, table=name)
        self.key = key
        self.description = description
        # todo: use last_updated_timestamp to match the record in metadata
        self.last_updated_timestamp = int(last_updated_timestamp) if last_updated_timestamp else None
        self.column_names = column_names
        self.column_descriptions = column_descriptions
        self.total_usage = total_usage
        self.unique_usage = unique_usage
        # todo: will include tag_type once we have better understanding from UI flow.
        self.tags = tags
        self.badges = badges
        self.schema_description = schema_description
        self.programmatic_descriptions = programmatic_descriptions



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import abc

import logging
import textwrap

LOGGER = logging.getLogger(__name__)


class RelationPreprocessor(object, metaclass=abc.ABCMeta):
    """
    A Preprocessor for relations. Prior to publish Neo4j relations, RelationPreprocessor will be used for
    pre-processing.
    Neo4j Publisher will iterate through relation file and call preprocess_cypher to perform any pre-process requested.

    For example, if you need current job's relation data to be desired state, you can add delete statement in
    pre-process_cypher method. With preprocess_cypher defined, and with long transaction size, Neo4j publisher will
    atomically apply desired state.


    """

    def preprocess_cypher(self,
                          start_label,
                          end_label,
                          start_key,
                          end_key,
                          relation,
                          reverse_relation):
        # type: (str, str, str, str, str, str) -> Tuple[str, Dict[str, str]]
        """
        Provides a Cypher statement that will be executed before publishing relations.
        :param start_label:
        :param end_label:
        :param start_key:
        :param end_key:
        :param relation:
        :param reverse_relation:
        :return:
        """
        if self.filter(start_label=start_label,
                       end_label=end_label,
                       start_key=start_key,
                       end_key=end_key,
                       relation=relation,
                       reverse_relation=reverse_relation):
            return self.preprocess_cypher_impl(start_label=start_label,
                                               end_label=end_label,
                                               start_key=start_key,
                                               end_key=end_key,
                                               relation=relation,
                                               reverse_relation=reverse_relation)

    @abc.abstractmethod
    def preprocess_cypher_impl(self,
                               start_label,
                               end_label,
                               start_key,
                               end_key,
                               relation,
                               reverse_relation):
        # type: (str, str, str, str, str, str) -> Tuple[str, Dict[str, str]]
        """
        Provides a Cypher statement that will be executed before publishing relations.
        :param start_label:
        :param end_label:
        :param relation:
        :param reverse_relation:
        :return: A Cypher statement
        """
        pass

    def filter(self,
               start_label,
               end_label,
               start_key,
               end_key,
               relation,
               reverse_relation):
        # type: (str, str, str, str, str, str) -> bool
        """
        A method that filters pre-processing in record level. Returns True if it needs preprocessing, otherwise False.
        :param start_label:
        :param end_label:
        :param start_key:
        :param end_key:
        :param relation:
        :param reverse_relation:
        :return: bool. True if it needs preprocessing, otherwise False.
        """
        True

    @abc.abstractmethod
    def is_perform_preprocess(self):
        # type: () -> bool
        """
        A method for Neo4j Publisher to determine whether to perform pre-processing or not. Regard this method as a
        global filter.
        :return: True if you want to enable the pre-processing.
        """
        pass


class NoopRelationPreprocessor(RelationPreprocessor):

    def preprocess_cypher_impl(self,
                               start_label,
                               end_label,
                               start_key,
                               end_key,
                               relation,
                               reverse_relation):
        # type: (str, str, str, str, str, str) -> Tuple[str, Dict[str, str]]
        pass

    def is_perform_preprocess(self):
        # type: () -> bool
        return False


class DeleteRelationPreprocessor(RelationPreprocessor):
    """
    A Relation Pre-processor that delete relationship before Neo4jPublisher publishes relations.

    Example use case: Take an example of an external privacy service trying to push personal identifiable
    identification (PII) tag into Amundsen. It is fine to push set of PII tags for the first push, but it becomes a
    challenge when it comes to following update as external service does not know current PII state in Amundsen.

    The easy solution is for external service to know desired state (certain columns should have certain PII tags),
    and push that information.
    Now the challenge is how Amundsen apply desired state. This is where DeleteRelationPreprocessor comes into the
    picture. We can utilize DeleteRelationPreprocessor to let it delete certain relations in the job,
    and let Neo4jPublisher update to desired state. Should there be a small window (between delete and update) that
    Amundsen data is not complete, you can increase Neo4jPublisher's transaction size to make it atomic. However,
    note that you should not set transaction size too big as Neo4j uses memory to store transaction and this use case
    is proper for small size of batch job.
    """
    RELATION_MERGE_TEMPLATE = textwrap.dedent("""
    MATCH (n1:{start_label} {{key: $start_key }})-[r]-(n2:{end_label} {{key: $end_key }})
    {where_clause}
    WITH r LIMIT 2
    DELETE r
    RETURN count(*) as count;
    """)

    def __init__(self, label_tuples=None, where_clause=''):
        # type: (List[Tuple[str, str]], str) -> None
        super(DeleteRelationPreprocessor, self).__init__()
        self._label_tuples = set(label_tuples) if label_tuples else set()

        reversed_label_tuples = [(t2, t1) for t1, t2 in self._label_tuples]
        self._label_tuples.update(reversed_label_tuples)
        self._where_clause = where_clause

    def preprocess_cypher_impl(self,
                               start_label,
                               end_label,
                               start_key,
                               end_key,
                               relation,
                               reverse_relation):
        # type: (str, str, str, str, str, str) -> Tuple[str, Dict[str, str]]
        """
        Provides DELETE Relation Cypher query on specific relation.
        :param start_label:
        :param end_label:
        :param start_key:
        :param end_key:
        :param relation:
        :param reverse_relation:
        :return:
        """

        if not (start_label or end_label or start_key or end_key):
            raise Exception('all labels and keys are required: {}'.format(locals()))

        params = {'start_key': start_key, 'end_key': end_key}
        return DeleteRelationPreprocessor.RELATION_MERGE_TEMPLATE.format(start_label=start_label,
                                                                         end_label=end_label,
                                                                         where_clause=self._where_clause), params

    def is_perform_preprocess(self):
        # type: () -> bool
        return True

    def filter(self,
               start_label,
               end_label,
               start_key,
               end_key,
               relation,
               reverse_relation):
        # type: (str, str, str, str, str, str) -> bool
        """
        If pair of labels is what client requested passed through label_tuples, filter will return True meaning that
        it needs to be pre-processed.
        :param start_label:
        :param end_label:
        :param start_key:
        :param end_key:
        :param relation:
        :param reverse_relation:
        :return: bool. True if it needs preprocessing, otherwise False.
        """
        if self._label_tuples and (start_label, end_label) not in self._label_tuples:
            return False

        return True



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
import re

from typing import Optional, Dict, Any, List, Union, Iterator  # noqa: F401

from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata
from databuilder.models.neo4j_csv_serde import (
    Neo4jCsvSerializable, RELATION_START_KEY, RELATION_END_KEY, RELATION_START_LABEL,
    RELATION_END_LABEL, RELATION_TYPE, RELATION_REVERSE_TYPE)
from databuilder.models.table_metadata import TableMetadata


LOGGER = logging.getLogger(__name__)


class DashboardTable(Neo4jCsvSerializable):
    """
    A model that link Dashboard with the tables used in various charts of the dashboard.
    Note that it does not create new dashboard, table as it has insufficient information but it builds relation
    between Tables and Dashboard
    """

    DASHBOARD_TABLE_RELATION_TYPE = 'DASHBOARD_WITH_TABLE'
    TABLE_DASHBOARD_RELATION_TYPE = 'TABLE_OF_DASHBOARD'

    def __init__(self,
                 dashboard_group_id,  # type: str
                 dashboard_id,  # type: str
                 table_ids,  # type: List[str]
                 product='',  # type: Optional[str]
                 cluster='gold',  # type: str
                 **kwargs
                 ):
        self._dashboard_group_id = dashboard_group_id
        self._dashboard_id = dashboard_id
        # A list of tables uri used in the dashboard
        self._table_ids = table_ids
        self._product = product
        self._cluster = cluster

        self._relation_iterator = self._create_relation_iterator()

    def create_next_node(self):
        # type: () -> Union[Dict[str, Any], None]
        return None

    def create_next_relation(self):
        # type: () -> Union[Dict[str, Any], None]
        try:
            return next(self._relation_iterator)
        except StopIteration:
            return None

    def _create_relation_iterator(self):
        # type: () -> Optional[None, Iterator[[Dict[str, Any]]]]
        for table_id in self._table_ids:
            m = re.match('(\w+)://(\w+)\.(\w+)\/(\w+)', table_id)
            if m:
                yield {
                    RELATION_START_LABEL: DashboardMetadata.DASHBOARD_NODE_LABEL,
                    RELATION_END_LABEL: TableMetadata.TABLE_NODE_LABEL,
                    RELATION_START_KEY: DashboardMetadata.DASHBOARD_KEY_FORMAT.format(
                        product=self._product,
                        cluster=self._cluster,
                        dashboard_group=self._dashboard_group_id,
                        dashboard_name=self._dashboard_id
                    ),
                    RELATION_END_KEY: TableMetadata.TABLE_KEY_FORMAT.format(
                        db=m.group(1),
                        cluster=m.group(2),
                        schema=m.group(3),
                        tbl=m.group(4)
                    ),
                    RELATION_TYPE: DashboardTable.DASHBOARD_TABLE_RELATION_TYPE,
                    RELATION_REVERSE_TYPE: DashboardTable.TABLE_DASHBOARD_RELATION_TYPE
                }

    def __repr__(self):
        return 'DashboardTable({!r}, {!r}, {!r}, {!r}, ({!r}))'.format(
            self._dashboard_group_id,
            self._dashboard_id,
            self._product,
            self._cluster,
            ','.join(self._table_ids),
        )



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Dict, List, Union  # noqa: F401

from databuilder.models.neo4j_csv_serde import Neo4jCsvSerializable, NODE_KEY, NODE_LABEL


class Neo4jESLastUpdated(Neo4jCsvSerializable):
    # type: (...) -> None
    """
    Data model to keep track the last updated timestamp for
    neo4j and es.
    """

    LABEL = 'Updatedtimestamp'
    KEY = 'amundsen_updated_timestamp'
    LATEST_TIMESTAMP = 'latest_timestmap'

    def __init__(self,
                 timestamp,  # type: int
                 ):
        # type: (...) -> None
        """
        :param timestamp: epoch for latest updated timestamp for neo4j an es
        """
        self.timestamp = timestamp
        self._node_iter = iter(self.create_nodes())
        self._rel_iter = iter(self.create_relation())

    def create_next_node(self):
        # type: (...) -> Union[Dict[str, Any], None]
        """
        Will create an orphan node for last updated timestamp.
        :return:
        """
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_nodes(self):
        # type: () -> List[Dict[str, Any]]
        """
        Create a list of Neo4j node records.
        :return:
        """
        return [{
            NODE_KEY: Neo4jESLastUpdated.KEY,
            NODE_LABEL: Neo4jESLastUpdated.LABEL,
            Neo4jESLastUpdated.LATEST_TIMESTAMP: self.timestamp
        }]

    def create_next_relation(self):
        # type: () -> Union[Dict[str, Any], None]
        """
        :return:
        """
        try:
            return next(self._rel_iter)
        except StopIteration:
            return None

    def create_relation(self):
        # type: () -> List[Dict[str, Any]]
        return []



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime  # noqa: F401


class FileMetadata(object):

    def __init__(self,
                 path,  # type: str
                 last_updated,  # type: datetime
                 size  # type: int
                 ):
        # type: (...) -> None
        self.path = path
        self.last_updated = last_updated
        self.size = size

    def __repr__(self):
        # type: () -> str
        return """FileMetadata(path={!r}, last_updated={!r}, size={!r})""" \
            .format(self.path, self.last_updated, self.size)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import abc

from pyhocon import ConfigTree  # noqa: F401

from databuilder import Scoped
from databuilder.callback import call_back
from databuilder.callback.call_back import Callback  # noqa: F401


class Publisher(Scoped):
    """
    A Publisher that writes dataset (not a record) in Atomic manner,
    if possible.
    (Either success or fail, no partial state)
    Use case: If you want to use neo4j import util or Load CSV util,
    that takes CSV file to load database, you need to first create CSV file.
    CSV file holds number of records, and loader can writes multiple records
    to it. Once loader finishes writing CSV file, you have complete CSV file,
    ready to publish to Neo4j. Publisher can take the location of CSV file,
    and push to Neo4j.

    """

    def __init__(self):
        self.call_backs = []  # type: List[Callback]

    @abc.abstractmethod
    def init(self, conf):
        # type: (ConfigTree) -> None
        pass

    def publish(self):
        try:
            self.publish_impl()
        except Exception as e:
            call_back.notify_callbacks(self.call_backs, is_success=False)
            raise e
        call_back.notify_callbacks(self.call_backs, is_success=True)

    @abc.abstractmethod
    def publish_impl(self):
        # type: () -> None
        """
        An implementation of publish method. Subclass of publisher is expected to write publish logic by overriding
        this method
        :return: None
        """
        pass

    def register_call_back(self, callback):
        # type: (Callback) -> None
        """
        Register any callback method that needs to be notified when publisher is either able to successfully publish
        or failed to publish
        :param callback:
        :return: None
        """
        self.call_backs.append(callback)

    def get_scope(self):
        # type: () -> str
        return 'publisher'


class NoopPublisher(Publisher):
    def __init__(self):
        # type: () -> None
        super(NoopPublisher, self).__init__()

    def init(self, conf):
        # type: (ConfigTree) -> None
        pass

    def publish_impl(self):
        # type: () -> None
        pass

    def get_scope(self):
        # type: () -> str
        return 'publisher.noop'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import copy
from typing import Union, Dict, Any  # noqa: F401

from databuilder.models.neo4j_csv_serde import Neo4jCsvSerializable, NODE_KEY, \
    NODE_LABEL, RELATION_START_KEY, RELATION_START_LABEL, RELATION_END_KEY, \
    RELATION_END_LABEL, RELATION_TYPE, RELATION_REVERSE_TYPE
from databuilder.publisher.neo4j_csv_publisher import UNQUOTED_SUFFIX


class User(Neo4jCsvSerializable):
    # type: (...) -> None
    """
    User model. This model doesn't define any relationship.
    """
    USER_NODE_LABEL = 'User'
    USER_NODE_KEY_FORMAT = '{email}'
    USER_NODE_EMAIL = 'email'
    USER_NODE_FIRST_NAME = 'first_name'
    USER_NODE_LAST_NAME = 'last_name'
    USER_NODE_FULL_NAME = 'full_name'
    USER_NODE_GITHUB_NAME = 'github_username'
    USER_NODE_TEAM = 'team_name'
    USER_NODE_EMPLOYEE_TYPE = 'employee_type'
    USER_NODE_MANAGER_EMAIL = 'manager_email'
    USER_NODE_SLACK_ID = 'slack_id'
    USER_NODE_IS_ACTIVE = 'is_active{}'.format(UNQUOTED_SUFFIX)  # bool value needs to be unquoted when publish to neo4j
    USER_NODE_UPDATED_AT = 'updated_at'
    USER_NODE_ROLE_NAME = 'role_name'

    USER_MANAGER_RELATION_TYPE = 'MANAGE_BY'
    MANAGER_USER_RELATION_TYPE = 'MANAGE'

    def __init__(self,
                 email,  # type: str
                 first_name='',  # type: str
                 last_name='',  # type: str
                 name='',  # type: str
                 github_username='',  # type: str
                 team_name='',  # type: str
                 employee_type='',  # type: str
                 manager_email='',  # type: str
                 slack_id='',  # type: str
                 is_active=True,  # type: bool
                 updated_at=0,  # type: int
                 role_name='',  # type: str
                 do_not_update_empty_attribute=False,  # type: bool
                 **kwargs  # type: Dict
                 ):
        # type: (...) -> None
        """
        This class models user node for Amundsen people.

        :param first_name:
        :param last_name:
        :param name:
        :param email:
        :param github_username:
        :param team_name:
        :param employee_type:
        :param manager_email:
        :param is_active:
        :param updated_at: everytime we update the node, we will push the timestamp.
                           then we will have a cron job to update the ex-employee nodes based on
                           the case if this timestamp hasn't been updated for two weeks.
        :param role_name: the role_name of the user (e.g swe)
        :param do_not_update_empty_attribute: If False, all empty or not defined params will be overwritten with
        empty string.
        :param kwargs: Any K/V attributes we want to update the
        """
        self.first_name = first_name
        self.last_name = last_name
        self.name = name

        self.email = email
        self.github_username = github_username
        # todo: team will be a separate node once Amundsen People supports team
        self.team_name = team_name
        self.manager_email = manager_email
        self.employee_type = employee_type
        # this attr not available in team service, either update team service, update with FE
        self.slack_id = slack_id
        self.is_active = is_active
        self.updated_at = updated_at
        self.role_name = role_name
        self.do_not_update_empty_attribute = do_not_update_empty_attribute
        self.attrs = None
        if kwargs:
            self.attrs = copy.deepcopy(kwargs)

        self._node_iter = iter(self.create_nodes())
        self._rel_iter = iter(self.create_relation())

    def create_next_node(self):
        # type: (...) -> Union[Dict[str, Any], None]
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self):
        # type: () -> Union[Dict[str, Any], None]
        """
        :return:
        """
        try:
            return next(self._rel_iter)
        except StopIteration:
            return None

    @classmethod
    def get_user_model_key(cls,
                           email=None):
        # type: (...) -> str
        if not email:
            return ''
        return User.USER_NODE_KEY_FORMAT.format(email=email)

    def create_nodes(self):
        # type: () -> List[Dict[str, Any]]
        """
        Create a list of Neo4j node records
        :return:
        """
        result_node = {
            NODE_KEY: User.get_user_model_key(email=self.email),
            NODE_LABEL: User.USER_NODE_LABEL,
            User.USER_NODE_EMAIL: self.email,
            User.USER_NODE_IS_ACTIVE: self.is_active,
        }

        result_node[User.USER_NODE_FIRST_NAME] = self.first_name if self.first_name else ''
        result_node[User.USER_NODE_LAST_NAME] = self.last_name if self.last_name else ''
        result_node[User.USER_NODE_FULL_NAME] = self.name if self.name else ''
        result_node[User.USER_NODE_GITHUB_NAME] = self.github_username if self.github_username else ''
        result_node[User.USER_NODE_TEAM] = self.team_name if self.team_name else ''
        result_node[User.USER_NODE_EMPLOYEE_TYPE] = self.employee_type if self.employee_type else ''
        result_node[User.USER_NODE_SLACK_ID] = self.slack_id if self.slack_id else ''
        result_node[User.USER_NODE_ROLE_NAME] = self.role_name if self.role_name else ''

        if self.updated_at:
            result_node[User.USER_NODE_UPDATED_AT] = self.updated_at
        elif not self.do_not_update_empty_attribute:
            result_node[User.USER_NODE_UPDATED_AT] = 0

        if self.attrs:
            for k, v in self.attrs.items():
                if k not in result_node:
                    result_node[k] = v

        if self.do_not_update_empty_attribute:
            for k, v in list(result_node.items()):
                if not v:
                    del result_node[k]

        return [result_node]

    def create_relation(self):
        # type: () -> List[Dict[str, Any]]
        if self.manager_email:
            # only create the relation if the manager exists
            return [{
                RELATION_START_KEY: User.get_user_model_key(email=self.email),
                RELATION_START_LABEL: User.USER_NODE_LABEL,
                RELATION_END_KEY: self.get_user_model_key(email=self.manager_email),
                RELATION_END_LABEL: User.USER_NODE_LABEL,
                RELATION_TYPE: User.USER_MANAGER_RELATION_TYPE,
                RELATION_REVERSE_TYPE: User.MANAGER_USER_RELATION_TYPE
            }]
        return []

    def __repr__(self):
        # type: () -> str
        return 'User({!r}, {!r}, {!r}, {!r}, {!r}, ' \
               '{!r}, {!r}, {!r}, {!r}, {!r}, {!r}, {!r})'.format(self.first_name,
                                                                  self.last_name,
                                                                  self.name,
                                                                  self.email,
                                                                  self.github_username,
                                                                  self.team_name,
                                                                  self.slack_id,
                                                                  self.manager_email,
                                                                  self.employee_type,
                                                                  self.is_active,
                                                                  self.updated_at,
                                                                  self.role_name)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging

from pyhocon import ConfigTree, ConfigFactory  # noqa: F401
from typing import Any  # noqa: F401

from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.dashboard.mode_analytics.mode_dashboard_utils import ModeDashboardUtils
from databuilder.rest_api.mode_analytics.mode_paginated_rest_api_query import ModePaginatedRestApiQuery
from databuilder.rest_api.rest_api_query import RestApiQuery
from databuilder.transformer.base_transformer import ChainedTransformer
from databuilder.transformer.dict_to_model import DictToModel, MODEL_CLASS
from databuilder.transformer.regex_str_replace_transformer import RegexStrReplaceTransformer, \
    REGEX_REPLACE_TUPLE_LIST, ATTRIBUTE_NAME
from databuilder.transformer.template_variable_substitution_transformer import \
    TemplateVariableSubstitutionTransformer, TEMPLATE, FIELD_NAME

LOGGER = logging.getLogger(__name__)


class ModeDashboardQueriesExtractor(Extractor):
    """
    A Extractor that extracts Query information

    """

    def init(self, conf):
        # type: (ConfigTree) -> None
        self._conf = conf

        restapi_query = self._build_restapi_query()
        self._extractor = ModeDashboardUtils.create_mode_rest_api_extractor(
            restapi_query=restapi_query,
            conf=self._conf
        )

        # Constructing URL using several ID via TemplateVariableSubstitutionTransformer
        transformers = []
        variable_substitution_transformer = TemplateVariableSubstitutionTransformer()
        variable_substitution_transformer.init(
            conf=Scoped.get_scoped_conf(self._conf,
                                        variable_substitution_transformer.get_scope()).with_fallback(
                ConfigFactory.from_dict({FIELD_NAME: 'url',
                                         TEMPLATE: 'https://app.mode.com/{organization}'
                                                   '/reports/{dashboard_id}/queries/{query_id}'})))

        transformers.append(variable_substitution_transformer)

        # Escape backslash as it breaks Cypher statement.
        replace_transformer = RegexStrReplaceTransformer()
        replace_transformer.init(
            conf=Scoped.get_scoped_conf(self._conf, replace_transformer.get_scope()).with_fallback(
                ConfigFactory.from_dict(
                    {REGEX_REPLACE_TUPLE_LIST: [('\\', '\\\\')], ATTRIBUTE_NAME: 'query_text'})))
        transformers.append(replace_transformer)

        dict_to_model_transformer = DictToModel()
        dict_to_model_transformer.init(
            conf=Scoped.get_scoped_conf(self._conf, dict_to_model_transformer.get_scope()).with_fallback(
                ConfigFactory.from_dict(
                    {MODEL_CLASS: 'databuilder.models.dashboard.dashboard_query.DashboardQuery'})))
        transformers.append(dict_to_model_transformer)

        self._transformer = ChainedTransformer(transformers=transformers)

    def extract(self):
        # type: () -> Any

        record = self._extractor.extract()
        if not record:
            return None

        return self._transformer.transform(record=record)

    def get_scope(self):
        # type: () -> str
        return 'extractor.mode_dashboard_query'

    def _build_restapi_query(self):
        """
        Build REST API Query. To get Mode Dashboard last execution, it needs to call three APIs (spaces API, reports
        API, and queries API) joining together.
        :return: A RestApiQuery that provides Mode Dashboard execution (run)
        """
        # type: () -> RestApiQuery

        spaces_query = ModeDashboardUtils.get_spaces_query_api(conf=self._conf)
        params = ModeDashboardUtils.get_auth_params(conf=self._conf)

        # Reports
        # https://mode.com/developer/api-reference/analytics/reports/#listReportsInSpace
        url = 'https://app.mode.com/api/{organization}/spaces/{dashboard_group_id}/reports'
        json_path = '(_embedded.reports[*].token)'
        field_names = ['dashboard_id']
        reports_query = ModePaginatedRestApiQuery(query_to_join=spaces_query, url=url, params=params,
                                                  json_path=json_path, field_names=field_names, skip_no_result=True)

        queries_url_template = 'https://app.mode.com/api/{organization}/reports/{dashboard_id}/queries'
        json_path = '_embedded.queries[*].[token,name,raw_query]'
        field_names = ['query_id', 'query_name', 'query_text']
        query_names_query = RestApiQuery(query_to_join=reports_query, url=queries_url_template, params=params,
                                         json_path=json_path, field_names=field_names, skip_no_result=True)

        return query_names_query



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging

from typing import Optional, Dict, Any, Union, Iterator  # noqa: F401

from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata
from databuilder.models.neo4j_csv_serde import (
    Neo4jCsvSerializable, NODE_LABEL, NODE_KEY, RELATION_START_KEY, RELATION_END_KEY, RELATION_START_LABEL,
    RELATION_END_LABEL, RELATION_TYPE, RELATION_REVERSE_TYPE)

LOGGER = logging.getLogger(__name__)


class DashboardExecution(Neo4jCsvSerializable):
    """
    A model that encapsulate Dashboard's execution timestamp in epoch and execution state
    """
    DASHBOARD_EXECUTION_LABEL = 'Execution'
    DASHBOARD_EXECUTION_KEY_FORMAT = '{product}_dashboard://{cluster}.{dashboard_group_id}/' \
                                     '{dashboard_id}/execution/{execution_id}'
    DASHBOARD_EXECUTION_RELATION_TYPE = 'EXECUTED'
    EXECUTION_DASHBOARD_RELATION_TYPE = 'EXECUTION_OF'

    LAST_EXECUTION_ID = '_last_execution'
    LAST_SUCCESSFUL_EXECUTION_ID = '_last_successful_execution'

    def __init__(self,
                 dashboard_group_id,  # type: Optional[str]
                 dashboard_id,  # type: Optional[str]
                 execution_timestamp,  # type: int
                 execution_state,  # type: str
                 execution_id=LAST_EXECUTION_ID,  # type: str
                 product='',  # type: Optional[str]
                 cluster='gold',  # type: str
                 **kwargs
                 ):
        self._dashboard_group_id = dashboard_group_id
        self._dashboard_id = dashboard_id
        self._execution_timestamp = execution_timestamp
        self._execution_state = execution_state
        self._execution_id = execution_id
        self._product = product
        self._cluster = cluster
        self._node_iterator = self._create_node_iterator()
        self._relation_iterator = self._create_relation_iterator()

    def create_next_node(self):
        # type: () -> Union[Dict[str, Any], None]
        try:
            return next(self._node_iterator)
        except StopIteration:
            return None

    def _create_node_iterator(self):  # noqa: C901
        # type: () -> Iterator[[Dict[str, Any]]]
        yield {
            NODE_LABEL: DashboardExecution.DASHBOARD_EXECUTION_LABEL,
            NODE_KEY: self._get_last_execution_node_key(),
            'timestamp': self._execution_timestamp,
            'state': self._execution_state
        }

    def create_next_relation(self):
        # type: () -> Union[Dict[str, Any], None]
        try:
            return next(self._relation_iterator)
        except StopIteration:
            return None

    def _create_relation_iterator(self):
        # type: () -> Iterator[[Dict[str, Any]]]
        yield {
            RELATION_START_LABEL: DashboardMetadata.DASHBOARD_NODE_LABEL,
            RELATION_END_LABEL: DashboardExecution.DASHBOARD_EXECUTION_LABEL,
            RELATION_START_KEY: DashboardMetadata.DASHBOARD_KEY_FORMAT.format(
                product=self._product,
                cluster=self._cluster,
                dashboard_group=self._dashboard_group_id,
                dashboard_name=self._dashboard_id
            ),
            RELATION_END_KEY: self._get_last_execution_node_key(),
            RELATION_TYPE: DashboardExecution.DASHBOARD_EXECUTION_RELATION_TYPE,
            RELATION_REVERSE_TYPE: DashboardExecution.EXECUTION_DASHBOARD_RELATION_TYPE
        }

    def _get_last_execution_node_key(self):
        return DashboardExecution.DASHBOARD_EXECUTION_KEY_FORMAT.format(
            product=self._product,
            cluster=self._cluster,
            dashboard_group_id=self._dashboard_group_id,
            dashboard_id=self._dashboard_id,
            execution_id=self._execution_id
        )

    def __repr__(self):
        return 'DashboardExecution({!r}, {!r}, {!r}, {!r}, {!r}, {!r}, {!r})'.format(
            self._dashboard_group_id,
            self._dashboard_id,
            self._execution_timestamp,
            self._execution_state,
            self._execution_id,
            self._product,
            self._cluster
        )



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import abc

from pyhocon import ConfigTree, ConfigFactory  # noqa: F401


class Scoped(object, metaclass=abc.ABCMeta):
    _EMPTY_CONFIG = ConfigFactory.from_dict({})
    """
    An interface for class that works with scoped (nested) config.
    https://github.com/chimpler/pyhocon
    A scoped instance will use config within its scope. This is a way to
    distribute configuration to its implementation instead of having it in
    one central place.
    This is very useful for DataBuilder as it has different components
    (extractor, transformer, loader, publisher) and its component itself
    could have different implementation.
    For example these can be a configuration for two different extractors
     "extractor.mysql.url" for MySQLExtractor
     "extractor.filesystem.source_path" for FileSystemExtractor

     For MySQLExtractor, if you defined scope as "extractor.mysql", scoped
     config will basically reduce it to the config that is only for MySQL.
     config.get("extractor.mysql") provides you all the config within
     'extractor.mysql'. By removing outer context from the config,
     MySQLExtractor is highly reusable.
    """

    @abc.abstractmethod
    def init(self, conf):
        # type: (ConfigTree) -> None
        """
        All scoped instance is expected to be lazily initialized. Means that
        __init__ should not have any heavy operation such as service call.
        The reason behind is that Databuilder is a code at the same time,
        code itself is used as a configuration. For example, you can
        instantiate scoped instance with all the parameters already set,
        ready to run, and actual execution will be executing init() and
        execute.

        :param conf: Typesafe config instance
        :return: None
        """
        pass

    @abc.abstractmethod
    def get_scope(self):
        # type: () -> str
        """
        A scope for the config. Typesafe config supports nested config.
        Scope, string, is used to basically peel off nested config
        :return:
        """
        return ''

    def close(self):
        # type: () -> None
        """
        Anything that needs to be cleaned up after the use of the instance.
        :return: None
        """
        pass

    @classmethod
    def get_scoped_conf(cls, conf, scope):
        # type: (ConfigTree, str) -> ConfigTree
        """
        Convenient method to provide scoped method.

        :param conf: Type safe config instance
        :param scope: scope string
        :return: Type safe config instance
        """
        if not scope:
            return Scoped._EMPTY_CONFIG

        return conf.get(scope, Scoped._EMPTY_CONFIG)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import re
from typing import Any, Dict, List, Union  # noqa: F401

from databuilder.models.neo4j_csv_serde import Neo4jCsvSerializable, \
    RELATION_START_KEY, RELATION_START_LABEL, RELATION_END_KEY, \
    RELATION_END_LABEL, RELATION_TYPE, RELATION_REVERSE_TYPE

from databuilder.models.table_metadata import TableMetadata


class TableLineage(Neo4jCsvSerializable):
    # type: (...) -> None
    """
    Table Lineage Model. It won't create nodes but create upstream/downstream rels.
    """
    LABEL = 'Lineage'
    KEY_FORMAT = '{db}://{cluster}.{schema}/{tbl}/'
    ORIGIN_DEPENDENCY_RELATION_TYPE = 'UPSTREAM'
    DEPENDENCY_ORIGIN_RELATION_TYPE = 'DOWNSTREAM'

    def __init__(self,
                 db_name,  # type: str
                 schema,  # type: str
                 table_name,  # type: str
                 cluster,  # type: str
                 downstream_deps=None,  # type: List
                 ):
        # type: (...) -> None
        self.db = db_name.lower()
        self.schema = schema.lower()
        self.table = table_name.lower()

        self.cluster = cluster.lower() if cluster else 'gold'
        # a list of downstream dependencies, each of which will follow
        # the same key
        self.downstream_deps = downstream_deps
        self._node_iter = iter(self.create_nodes())
        self._relation_iter = iter(self.create_relation())

    def create_next_node(self):
        # type: (...) -> Union[Dict[str, Any], None]
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self):
        # type: (...) -> Union[Dict[str, Any], None]
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def get_table_model_key(self, db, cluster, schema, table):
        # type: (...) -> str
        return '{db}://{cluster}.{schema}/{table}'.format(db=db,
                                                          cluster=cluster,
                                                          schema=schema,
                                                          table=table)

    def create_nodes(self):
        # type: () -> List[Union[Dict[str, Any], None]]
        """
        It won't create any node for this model
        :return:
        """
        return []

    def create_relation(self):
        # type: () -> List[Dict[str, Any]]
        """
        Create a list of relation between source table and all the downstream tables
        :return:
        """
        results = []
        for downstream_tab in self.downstream_deps:
            # every deps should follow '{db}://{cluster}.{schema}/{table}'
            # todo: if we change the table uri, we should change here.
            m = re.match('(\w+)://(\w+)\.(\w+)\/(\w+)', downstream_tab)
            if m:
                # if not match, skip those records
                results.append({
                    RELATION_START_KEY: self.get_table_model_key(db=self.db,
                                                                 cluster=self.cluster,
                                                                 schema=self.schema,
                                                                 table=self.table),
                    RELATION_START_LABEL: TableMetadata.TABLE_NODE_LABEL,
                    RELATION_END_KEY: self.get_table_model_key(db=m.group(1),
                                                               cluster=m.group(2),
                                                               schema=m.group(3),
                                                               table=m.group(4)),
                    RELATION_END_LABEL: TableMetadata.TABLE_NODE_LABEL,
                    RELATION_TYPE: TableLineage.ORIGIN_DEPENDENCY_RELATION_TYPE,
                    RELATION_REVERSE_TYPE: TableLineage.DEPENDENCY_ORIGIN_RELATION_TYPE
                })
        return results

    def __repr__(self):
        # type: () -> str
        return 'TableLineage({!r}, {!r}, {!r}, {!r})'.format(self.db,
                                                             self.cluster,
                                                             self.schema,
                                                             self.table)
